PDGind = {
    0     : ('Total',                 'Total'),
#COMMON
    2212  : ('Proton',                'p'),
    -2212 : ('Antiproton',            r'$\overline{p}$'),
    2112  : ('Neutron',               'n'),
    -2112 : ('AntiNeutron',           r'$\overline{n}$'),
    11    : ('Electron',              r'$e^{-}$'),
    -11   : ('Positron',              r'$e^{+}$'),
    22    : ('Photon',                r'$\gamma$'),
    12    : ('Electron Neutrino',     r'$\nu_{e}$'),
    -12   : ('Electron Antineutrino', r'$\overline\nu_{e}$'),
    13    : ('Muon',                  r'$\mu^{-}$'),
    -13   : ('Antimuon',              r'$\mu^{+}$'),
    14    : ('Muon Neutrino',         r'$\nu_{\mu}$'),
    -14   : ('Muon Antineutrino',     r'$\overline\nu_{\mu}$'),
    15    : ('Tau',                   r'$\tau^{-}$'),
    -15   : ('Anti Tau',              r'$\tau^{+}$'),
    16    : ('Tau Neutrino',          r'$\nu_{\tau}$'),
    -16   : ('Tau Antineutrino ',     r'$\overline\nu_{\tau}$'),
#LIGHT MESONS
    111   : ('Pion0',                 r'$\pi^{0}$'),
    211   : ('Pion+',                 r'$\pi^{+}$'),
    -211  : ('Pion-',                 r'$\pi^{-}$'),
#STRANGE MESONS
    321   : ('Kaon+',                 r'$K^{+}$'),
    -321  : ('Kaon-',                 r'$K^{-}$'),
    130   : ('K-Long',                r'$K_{L}^{0}$'),
    310   : ('K-Short',               r'$K_{S}^{0}$'),
#STRANGE BARYONS
    3122  : ('Lambda',                r'$\Lambda$'),
    -3122 : ('Anti Lambda',           r'$\overline{\Lambda}$'),
    3222  : ('Sigma+',                r'$\Sigma^{+}$'),
    -3222 : ('Anti Sigma+',           r'$\overline{\Sigma^{+}}$'),
    3212  : ('Sigma0',                r'$\Sigma^{0}$'),
    3112  : ('Sigma-',                r'$\Sigma^{-}$'),
    -3112 : ('Anti Sigma-',           r'$\overline{\Sigma^{-}}$'),
    3322  : ('Xi0',                   r'$\Xi^{0}$'),
    -3322 : ('Anti Xi0',              r'$\overline{\Xi^{0}}$'),
    3312  : ('Xi-',                   r'$\Xi^{-}$'),
    -3312 : ('Anti Xi-',              r'$\overline{\Xi^{-}}$')
}

PDGname = {}
for k,v in list(PDGind.items()):
    PDGname[v[0]] = k
    PDGname[v[0].lower()] = k
del k,v

def GetPDGInd(particlename):
    if particlename in PDGname:
        return PDGname[particlename]
    elif particlename.lower() in PDGname:
        return PDGname[particlename.lower()]
    else:
        raise ValueError("Unknown particle type")

def GetPDGName(particleid):
    try:
        return PDGind[particleid]
    except KeyError:
        print('Unknown particle id ',particleid)
        return ('','')



import gzip as _gzip
import numpy as _np
import tarfile as _tarfile
from copy import deepcopy as _deepcopy

class Field(object):
    """
    Base class used for common writing procedures for BDSIM field format.

    This does not support arbitrary loop ordering - only the originally intended
    xyzt.
    """
    def __init__(self, array=_np.array([]), columns=[], flip=False, doublePrecision=False):
        self.data            = array
        self.columns         = columns
        self.header          = {}
        self.flip            = flip
        self.doublePrecision = doublePrecision
        self.nDimensions     = 0
        self.comments        = []

    def __add__(self, field):
        if self.nDimensions == field.nDimensions:
            self.data[..., self.nDimensions:] = self.data[..., self.nDimensions:] + field.data[..., field.nDimensions:]
            return self
        else:
            raise ValueError("The two field maps do not have the same dimension!")

    def __iadd__(self, field):
        self = self + field
        return self

    def __mul__(self, scalingFactor):
        self.data[..., self.nDimensions:] *= scalingFactor
        return self

    def __imul__(self, scalingFactor):
        self = self * scalingFactor
        return self

    def ScaleField(self, scalingFactor):
        self *= scalingFactor

    def AddField(self, field):
        self += field

    def AddComment(self, commentString):
        """
        Add a string that will be put on a comment line at the beginning of the file.
        """
        self.comments.append(str(commentString))
        
    def Write(self, fileName, writeLoopOrderReversed=False, overrideLoopOrder=None):
        """
        :param writeLoopOrderReversed: Write this field map with the other loop order.
        :type writeLoopOrderReversed: bool
        :param overrideLoopOrder: string to write irrespective of internal data as the loop order.
        :type overrideLoopOrder: str

        gzip - if the file ends with ".gz" the file will be compressed automatically.

        For overrideLoopOrder it should be only 'xyzt' or 'tzyx'. This option is
        provided in case a field is prepared in the other order somehow and you
        want to control the writing of this header variable independently.
        """
        compressed = fileName.endswith(".gz")
        if compressed:
            f = _gzip.open(fileName, 'wb')
        else:
            f = open(fileName, 'w')

        def write(fn, s):
            if compressed:
                fn.write(s.encode('ascii'))
            else:
                fn.write(s)
        write(f, "# units: cm, T\n")
        for comment in self.comments:
            write(f, "# "+str(comment).strip()+"\n")
        for key,value in self.header.items():
            write(f, str(key)+'> '+ str(value) + '\n')
        if overrideLoopOrder:
            if overrideLoopOrder not in ['xyzt', 'tzyx']:
                raise ValueError("overrideLoopOrder must be one of 'xyzt', 'tzyx'")
            write(f, "loopOrder> "+overrideLoopOrder+"\n")
        else:
            lo = 'tzyx' if writeLoopOrderReversed else 'xyzt'
            write(f, "loopOrder> "+lo+"\n")

        if self.doublePrecision:
            colStrings = ['%23s' % s for s in self.columns]
        else:
            colStrings = ['%14s' % s for s in self.columns]
        colStrings[0] = colStrings[0].strip() # don't pad the first column title
        # a '!' denotes the column header line
        write(f, '! '+ '\t'.join(colStrings)+'\n')
        
        # flatten all but last dimension - 3 field components
        nvalues = _np.shape(self.data)[-1] # number of values in last dimension

        flipLocal = self.flip
        if writeLoopOrderReversed:
            flipLocal = not flipLocal
        
        if not flipLocal:
            # [x,y,z,t,values] -> [t,z,y,x,values] for 4D
            # [x,y,z,values]   -> [z,y,x,values]   for 3D
            # [x,y,values]     -> [y,x,values]     for 2D
            # [x,values]       -> [x,values]       for 1D
            if (self.data.ndim == 2):
                pass # do nothin for 1D
            inds = list(range(self.data.ndim))       # indices for dimension [0,1,2] etc
            # keep the last value the same but reverse all indices before then
            inds[:(self.data.ndim - 1)] = reversed(inds[:(self.data.ndim - 1)])
            datal = _np.transpose(self.data, inds)
        else:
            datal = self.data

        datal = datal.reshape(-1,nvalues)
        for value in datal:
            if self.doublePrecision:
                strings   = ['%.16E' % x for x in value]
                stringsFW = ['%23s' % s for s in strings]
            else:
                strings   = ['%.8E' % x for x in value]
                stringsFW = ['%14s' % s for s in strings]
            write(f, '\t'.join(stringsFW) + '\n')

        f.close()

    def WriteFLUKA2DFormat1(self, fileName):
        """
        Write one of the FLUKA formats (x,y,Bx,by) in cm,T with no header.
    
        Very simple and only works for Field2D - can be improved.
        """
        if self.nDimensions != 2:
            raise ValueError("This field map is not 2D - it's ",self.nDimensions,"D")

        f = open(fileName, "w")
        
        # flatten all but last dimension - 3 field components
        nvalues = _np.shape(self.data)[-1] # number of values in last dimension

        flipLocal = self.flip

        dt = self.data.reshape(-1,nvalues)
        for xi in range(self.header['nx']):
            for yi in range(self.header['ny']):
                v = self.data[xi][yi]
                v = [v[index] for index in [0,1,2,3]] # [x,y,Bx,By]
                strings   = ['%.7E' % x for x in v]
                #stringsFW = ['%14s' % s for s in strings]
                f.write(','.join(strings) + '\n')
        
        f.close()

    def WriteMGNDataCard(self, fileName, name = 'magnet', symmetry = '0.0'):
        """
        :param fileName: Name of file to write the fieldmap to.
        :type fileName: str
        :param name: Name of the magnetic field in the MGNCREAT card.
        :type name: str
        :param symmetry: Symmetry that gets applied to the fieldmap inside FLUKA.
        :type symmetry: str

        Write 2D field maps stored in Field2D to FLUKA format which can be loaded with MGN data cards.
        Therefore, we create a magnet which is purely based on a fieldmap (200.0) without offsetting it, as
        (0,0) in the fieldmap is, where the beam is (and it's only used for analytical fields).
        The symmetry needs to be specified in accordance to Sx + Sy*10 + Sz*100, for example for a quadrupole
        it's 12.0 and for a C-shaped bend it's 10.0. 0.0 means no symmetry.
        In the final file, one can see three field points per MGNDATA card without specifying the location,
        as this is given by the grid definition. In the file, the points need to go from low x and y
        to high x and y, with looping first over y and then x.
        """
        if self.nDimensions != 2:
            raise ValueError("This field map is not 2D - it's ",self.nDimensions,"D")
        f = open(fileName, 'w')

        numberOfDigits = 9

        f.write('FREE\n')
        f.write('MGNCREAT  , 200.0, , 0.0, 0.0, ' + symmetry + ', , ' + name + '\n')
        f.write('MGNCREAT  , , , , ' + str(self.header['nx']) + ', ' + str(self.header['ny'])
                + ', ,  &\n')
        f.write('MGNCREAT  , ' +  str(self.header['xmin']) + ', ' + str(self.header['ymin'])
                + ', , ' + str(self.header['xmax']) + ', ' + str(self.header['ymax']) + ',  , &&\n')
        fieldPointCounter = 0 #Maximal 3 points per card
        lineCounter = 0 #First card ends with name of mgncreat card, second one with & and then &&
        line = 'MGNDATA   , '
        for yi in range(self.header['ny']):
            for xi in range(self.header['nx']):
                line += str(round(self.data[xi][yi][2], numberOfDigits)) + ', ' + str(round(self.data[xi][yi][3], numberOfDigits)) + ', '
                if fieldPointCounter == 2 and lineCounter == 0:
                    line += ' ' + name + '\n'
                    f.write(line)
                    line = 'MGNDATA   , '
                    lineCounter += 1
                    fieldPointCounter = 0
                elif fieldPointCounter == 2 and lineCounter == 1:
                    line += '  &\n'
                    f.write(line)
                    line = 'MGNDATA   , '
                    lineCounter += 1
                    fieldPointCounter = 0
                elif fieldPointCounter == 2 and lineCounter > 1:
                    line += '  &&\n'
                    f.write(line)
                    line = 'MGNDATA   , '
                    fieldPointCounter = 0
                elif fieldPointCounter < 2:
                    fieldPointCounter += 1

        if fieldPointCounter != 0:
            line += ', , ' * (3 - fieldPointCounter) + '  &&\nFIXED'
            f.write(line)
        else:
            f.write('FIXED')

class Field1D(Field):
    """
    Utility class to write a 1D field map array to BDSIM field format.

    The array supplied should be 2 dimensional. Dimensions are:
    (x,value) where value has 4 elements [x,fx,fy,fz]. So a 120 long
    array would have np.shape of (120,4).
    
    This can be used for both electric and magnetic fields.

    Example::
    
    >>> a = Field1D(data)
    >>> a.Write('outputFileName.dat')

    """
    def __init__(self, data, doublePrecision=False, column='X'):
        columns = [column,'Fx','Fy','Fz']
        super(Field1D, self).__init__(data,columns,doublePrecision=doublePrecision)
        self.header[column.lower() + 'min'] = _np.min(self.data[:,0])
        self.header[column.lower() + 'max'] = _np.max(self.data[:,0])
        self.header['n' + column.lower()]   = _np.shape(self.data)[0]
        self.nDimensions = 1


class Field2D(Field):
    """
    Utility class to write a 2D field map array to BDSIM field format.

    The array supplied should be 3 dimensional. Dimensions are:
    (x,y,value) where value has 5 elements [x,y,fx,fy,fz].  So a 100x50 (x,y)
    grid would have np.shape of (100,50,5).

    Example::
    
    >>> a = Field2D(data) # data is a prepared array
    >>> a.Write('outputFileName.dat')

    The 'flip' boolean allows an array with (y,x,value) dimension order
    to be written as (x,y,value). Values must still be (x,y,fx,fy,fz).

    The 'doublePrecision' boolean controls whether the field and spatial
    values are written to 16 s.f. (True) or 8 s.f. (False - default).

    """
    def __init__(self, data, flip=False, doublePrecision=False, firstColumn='X', secondColumn='Y'):
        columns = [firstColumn, secondColumn, 'Fx', 'Fy', 'Fz']
        super(Field2D, self).__init__(data,columns,flip,doublePrecision)
        inds = [1,0] if flip else [0,1]
        fcl = firstColumn.lower()
        scl = secondColumn.lower()
        self.header[fcl+'min'] = _np.min(self.data[:,:,0])
        self.header[fcl+'max'] = _np.max(self.data[:,:,0])
        self.header['n'+fcl]   = _np.shape(self.data)[inds[0]]
        self.header[scl+'min'] = _np.min(self.data[:,:,1])
        self.header[scl+'max'] = _np.max(self.data[:,:,1])
        self.header['n'+scl]   = _np.shape(self.data)[inds[1]]
        self.nDimensions = 2


class Field3D(Field):
    """
    Utility class to write a 3D field map array to BDSIM field format.

    The array supplied should be 4 dimensional. Dimensions are:
    (x,y,z,value) where value has 6 elements [x,y,z,fx,fy,fz].  So a 100x50x30 
    (x,y,z) grid would have np.shape of (100,50,30,6).
    
    Example::
    
    >>> a = Field3D(data) # data is a prepared array
    >>> a.Write('outputFileName.dat')

    The 'flip' boolean allows an array with (z,y,x,value) dimension order to
    be written as (x,y,z,value). Values must still be (x,y,fx,fy,fz).

    The 'doublePrecision' boolean controls whether the field and spatial
    values are written to 16 s.f. (True) or 8 s.f. (False - default).

    """
    def __init__(self, data, flip=False, doublePrecision=False, firstColumn='X', secondColumn='Y', thirdColumn='Z'):
        columns = [firstColumn,secondColumn,thirdColumn,'Fx','Fy','Fz']
        super(Field3D, self).__init__(data,columns,flip,doublePrecision)
        inds = [2,1,0] if flip else [0,1,2]
        fcl = firstColumn.lower()
        scl = secondColumn.lower()
        tcl = thirdColumn.lower()
        self.header[fcl+'min'] = _np.min(self.data[:,:,:,0])
        self.header[fcl+'max'] = _np.max(self.data[:,:,:,0])
        self.header['n'+fcl]   = _np.shape(self.data)[inds[0]]
        self.header[scl+'min'] = _np.min(self.data[:,:,:,1])
        self.header[scl+'max'] = _np.max(self.data[:,:,:,1])
        self.header['n'+scl]   = _np.shape(self.data)[inds[1]]
        self.header[tcl+'min'] = _np.min(self.data[:,:,:,2])
        self.header[tcl+'max'] = _np.max(self.data[:,:,:,2])
        self.header['n'+tcl]   = _np.shape(self.data)[inds[2]]
        self.nDimensions = 3


class Field4D(Field):
    """
    Utility class to write a 4D field map array to BDSIM field format.

    The array supplied should be 5 dimensional. Dimensions are:
    (t,y,z,x,value) where value has 7 elements [x,y,z,t,fx,fy,fz]. So a 100x50x30x10
    (x,y,z,t) grid would have np.shape of (10,30,50,100,7).
    
    Example::
    
    >>> a = Field4D(data) # data is a prepared array
    >>> a.Write('outputFileName.dat')

    The 'flip' boolean allows an array with (t,z,y,x,value) dimension order to
    be written as (x,y,z,t,value). Values must still be (x,y,fx,fy,fz).

    The 'doublePrecision' boolean controls whether the field and spatial
    values are written to 16 s.f. (True) or 8 s.f. (False - default).

    """
    def __init__(self, data, flip=False, doublePrecision=False):
        columns = ['X','Y','Z','T','Fx','Fy','Fz']
        super(Field4D, self).__init__(data,columns,flip,doublePrecision)
        inds = [3,2,1,0] if flip else [0,1,2,3]
        self.header['xmin'] = _np.min(self.data[:,:,:,:,0])
        self.header['xmax'] = _np.max(self.data[:,:,:,:,0])
        self.header['nx']   = _np.shape(self.data)[inds[0]]
        self.header['ymin'] = _np.min(self.data[:,:,:,:,1])
        self.header['ymax'] = _np.max(self.data[:,:,:,:,1])
        self.header['ny']   = _np.shape(self.data)[inds[1]]
        self.header['zmin'] = _np.min(self.data[:,:,:,:,2])
        self.header['zmax'] = _np.max(self.data[:,:,:,:,2])
        self.header['nz']   = _np.shape(self.data)[inds[2]]
        self.header['tmin'] = _np.min(self.data[:,:,:,:,3])
        self.header['tmax'] = _np.max(self.data[:,:,:,:,3])
        self.header['nt']   = _np.shape(self.data)[inds[3]]
        self.nDimensions = 4


def Load(filename, debug=False):
    """
    :param filename: name of file to load
    :type filename: str
    
    Load a BDSIM field format file into a numpy array. Can either
    be a regular ascii text file or can be a compressed file ending
    in ".tar.gz".

    returns a numpy array with the corresponding number of dimensions
    and the dimension has the coordinates and fx,fy,fz.
    """
    gzippedFile = False
    if (filename.endswith('.tar.gz')):
        if debug:
            print('Field Loader> loading compressed file ' + filename)
        tar = _tarfile.open(filename,'r')
        f = tar.extractfile(tar.firstmember)
    elif '.gz' in filename:
        f = _gzip.open(filename)
        gzippedFile = True
    else:
        if debug:
            print('Field Loader> loading file ' + filename)
        f = open(filename)

    intoData = False
    header   = {}
    columns  = []
    data     = []
    comments = []
    
    for line in f:
        if gzippedFile:
            line = line.decode('utf-8')
        if intoData:
            ls = line.strip()
            # avoid empty lines
            if ls.isspace() or len(ls) == 0:
                continue
            data.append(line.strip().split())
        elif '>' in line:
            d = line.strip().split('>')
            k = d[0].strip()
            try:
                v = float(d[1].strip())
            except ValueError:
                v = d[1].strip()
            header[k] = v
        elif '!' in line:
            columns = line.strip('!').strip().split()
            intoData = True
        elif line.lstrip().startswith('#'):
            comments.append(line.lstrip()[1:])

    f.close()
    
    data = _np.array(data, dtype=float)

    normalLoopOrder = ['x','y','z','t']
    # this is convention - in the case of xyzt, bdsim loops
    # over x first, then y, then z, so it appears the first
    # column is changing.
    flip = False

    if 'loopOrder' in header:
        order = header['loopOrder']
        flip = order == 'tzyx'
        if debug:
            print("flip :",flip)

    nDim = len(columns) - 3
    # TBC for no case when we don't store spatial coords also
    if (nDim < 1 or nDim > 4):
        if debug:
            print('Invalid number of columns')
            print(columns)
        return

    requiredSet = {'nx','ny','nz','nt'}
    headerKeySet = set(header.keys())
    keysPresent = headerKeySet.intersection(requiredSet)
    if len(keysPresent) < nDim:
        print('missing keys from header!')
        if debug:
            print(header)
        return
    else:
        dimToNVariable = {'x' : 'nx',
                          'y' : 'ny',
                          'z' : 'nz',
                          't' : 'nt'}
        if debug:
            print("Columns: ", columns)
            print("Header: ", header)
        dims = [int(header[dimToNVariable[k.lower()]]) for k in columns[:-3]]
        dims.append(len(columns))
        if debug:
            print("Shape of numpy array to be: ", dims)
            print("nDimensions: ", nDim)
            print("Existing numpy array shape: ", _np.shape(data))
        data = data.reshape(*dims)

    # build field object
    #columns = [s.strip('n') for s in keysPresentList]
    if nDim == 1:
        fd = Field1D(data, column=columns[0])
    elif nDim == 2:
        fd = Field2D(data, flip=flip, firstColumn=columns[0], secondColumn=columns[1])
    elif nDim == 3:
        fd = Field3D(data, flip=flip, firstColumn=columns[0], secondColumn=columns[1], thirdColumn=columns[2])
    elif nDim == 4:
        fd = Field4D(data, flip=flip)
    else:
        raise ValueError("Invalid number of dimensions")

    fd.comments = comments
    return fd


def MirrorDipoleQuadrant1(field2D):
    """
    +-------+-------+
    |       |       |
    |   2   |   1   |
    |       |       |
    +-------+-------+
    |       |       |
    |   3   |   4   |
    |       |       |
    +-------+-------+

    :param field2D: field object
    :type field2D: pybdsim.Field._Field.Field2D instance

    returns an instance of the same type.
    
    For a 2D field (i.e. function of x,y but can include Bx,By,Bz),
    for the quadrant number 1, mirror it and generate a bigger field for
    all four quadrants.
    
    1. original data
    2. data mirrored in x, (x,Bx) \*= -1
    3. data mirrored in x,y, (x,y,By) \*= -1
    4. data mirrored in y, (y,Bx) \*= -1

    This is based on a dipole field.
    """
    d = field2D.data
    ds = _np.shape(d)
    sx = ds[0]
    sy = ds[1]
    result = _np.empty((2*sx, 2*sy, ds[2]))
    # top right quadrant
    result[sx:,sy:,:] = d
    # top left quadrant
    result[:sx,sy:,:] = d[::-1,:,:]
    # bottom left quadrant
    result[:sx,:sy] = d[::-1,::-1,:]
    # bottom right quadrant
    result[sx:,:sy,:] = d[:,::-1,:]

    # each value is x,y,Bx,By,Bz
    
    # flip x,Bx for top left
    result[:sx,sy:,_np.array([0,2])] *= -1
    # flip x,y for bottom left
    result[:sx,:sy,:2] *= -1
    # flip y,Bx for bottom right
    result[sx:,:sy,_np.array([1,2])] *= -1
    
    resultField = Field2D(result)
    return resultField


def SortUnorderedFieldMap2D(field):
    """
    Rearrange the data in a 2D field map to be in a linearly
    progressing loop of (x,y). In future this could be generalised
    to more dimensions and also any two dimensions, not just x,y.

    :param field: Incoming jumbled field map
    :type  field: pybdsim.Field.Field2D

    Returns a new Field2D object
    """
    d  = field.data
    sh = _np.shape(d)
    d2 = fd.data.reshape((sh[0]*sh[1],sh[2]))

    # build a dict with key (x,y)
    dmap = {}
    for v in d2:
        dmap[(v[0],v[1])] = v[2:]

    # build up new array now in order
    dnew = []
    h = fd.header
    for y in np.linspace(h['ymin'], h['ymax'], h['ny']):
        r = []
        for x in np.linspace(h['xmin'], h['xmax'], h['nx']):
            fv = dmap[(x,y)]
            r.append([x,y,*fv])
        dnew.append(r)

    dnew = _np.array(dnew)
    fdnew = Field2D(dnew)
    return fdnew



import pysad as _pysad
import pybdsim as _pybdsim
import matplotlib.pyplot as _plt

def SadComparison(sad, bdsim, survey=None, functions=None, figsize=(12,5)):
    sad = _pysad.Reader.Flat(sad)
    bds = _pybdsim.Data.Load(bdsim)

    PlotAlphas(sad,bds)
    PlotDs(sad,bds)
    PlotDps(sad,bds)
    PlotSigmas(sad,bds)

def PlotSigmas(sad, bds, figsize=(12,5)):
    betaPlot = _plt.figure('Sigma', figsize=figsize)

    # sad
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['SIGMAX'], 'b', label=r'SAD $\sigma_{x}$')
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['SIGMAY'], 'g', label=r'SAD $\sigma_{y}$')

    # bds
    _plt.errorbar(bds.S(),
                  bds.Sigma_x(),
                  yerr=bds.Sigma_Sigma_x(),
                  label=r'BDSIM $\sigma_{x}$',
                  fmt='b.', capsize=3)

    # bds
    _plt.errorbar(bds.S(),
                  bds.Sigma_y(),
                  yerr=bds.Sigma_Sigma_y(),
                  label=r'BDSIM $\sigma_{y}$',
                  fmt='g.', capsize=3)


    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\sigma_{x,y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

def PlotDs(sad, bds, figsize=(12, 5)):
    dispPlot = _plt.figure('Dispersion', figsize=figsize)

    # sad
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['EX'], 'b', label=r'SAD $\eta_{x}$')
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['EY'], 'g', label=r'SAD $\eta_{y}$')

    # bds
    _plt.errorbar(bds.S(),
                  bds.Disp_x(),
                  yerr=bds.Sigma_Disp_x(),
                  label=r'BDSIM $\eta_{x}$',
                  fmt='b.', capsize=3)

    # bds
    _plt.errorbar(bds.S(),
                  bds.Sigma_y(),
                  yerr=bds.Sigma_Disp_y(),
                  label=r'BDSIM $\eta_{y}$',
                  fmt='g.', capsize=3)


    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\eta_{x,y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

def PlotDps(sad, bds, figsize=(12, 5)):
    dpsPlot = _plt.figure('Angular dispersion', figsize=figsize)

    # sad
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['EPX'], 'b', label=r'SAD $\eta_{x}$')
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['EPY'], 'g', label=r'SAD $\eta_{y}$')

    # bds
    _plt.errorbar(bds.S(),
                  bds.Disp_xp(),
                  yerr=bds.Sigma_Disp_xp(),
                  label=r'BDSIM $\eta_{x}^{\prime}$',
                  fmt='b.', capsize=3)

    # bds
    _plt.errorbar(bds.S(),
                  bds.Sigma_yp(),
                  yerr=bds.Sigma_Disp_yp(),
                  label=r'BDSIM $\eta_{y}^{\prime}$',
                  fmt='g.', capsize=3)


    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\eta_{x,y}^{\prime}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')


def PlotAlphas(sad, bds, figsize=(12, 5)):
    dispPlot = _plt.figure('Alpha', figsize=figsize)

    # sad
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['AX'], 'b', label=r'SAD $\alpha_{x}$')
    _plt.plot(sad.dataDict['POSITION'], sad.dataDict['AY'], 'g', label=r'SAD $\alpha_{y}$')

    # bds
    _plt.errorbar(bds.S(),
                  bds.Alpha_x(),
                  yerr=bds.Sigma_Alpha_x(),
                  label=r'BDSIM $\alpha_{x}$',
                  fmt='b.', capsize=3)

    # bds
    _plt.errorbar(bds.S(),
                  bds.Alpha_y(),
                  yerr=bds.Sigma_Alpha_y(),
                  label=r'BDSIM $\alpha_{y}$',
                  fmt='g.', capsize=3)


    axes = _plt.gcf().gca()
    axes.set_ylabel(r'$\alpha_{x,y}$ / m')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')



"""
ModelProcessing

Tools to process existing BDSIM models and generate other
versions of them.

"""

from . import _General
from . import Gmad as _Gmad
from . import Builder as _Builder

import time as _time


def GetAxisAlignedBoundingBoxOfCollimator(modelData, collimatorName):
    """
    Inspect pybdsim.Data.ModelData assuming collimator info was stored
    to give an axis aligned bounding box set of ranges in global coordinates.
    """
    import pybdsim as _pybdsim
    if type(modelData) is not _pybdsim.Data.ModelData:
        raise TypeError("not pybdsim.Data.ModelData instance")

    ci = modelData.collimatorInfo[collimatorName]
    gi = modelData.collimatorIndicesByName[collimatorName]

    midPos = modelData.midRefPos[gi]
    midRot = modelData.midRefRot[gi]

    localRanges = [[-0.1,0.1], [-0.1,0.1], [-ci.length*0.5, ci.length*0.5]]
    
    
def GetMaterialIDOfCollimator(modelData, collimatorName):
    """
    Inspect pybdsim.Data.ModelData assuming collimator info was stored
    to give an axis aligned bounding box set of ranges in global coordinates.
    """
    import pybdsim as _pybdsim
    if type(modelData) is not _pybdsim.Data.ModelData:
        raise TypeError("not pybdsim.Data.ModelData instance")
    ci = modelData.collimatorInfo[collimatorName]
    return modelData.materialNameToID[ci.material]    

def GenerateFullListOfSamplers(inputfile, outputfile):
    """
    inputfile - path to main gmad input file

    This will parse the input using the compiled BDSIM
    parser (GMAD), iterate over all the beamline elements
    and generate a sampler for every element.  Ignores
    samplers, but may include already defined ones in your
    own input.

    """
    lattice  = _Gmad.Lattice(inputfile)
    samplers = []
    typestoignore = [
        'None',
        'Sampler',
        'CSampler',
        'Line',
        'Reversed Line',
        'Material',
        'Atom',
        'Sequence',
        'Tunnel',
        'Teleporter',
        'Terminator',
        'Transform3D'
        ]
    for e in lattice:
        if e['Type'] not in typestoignore:
           samplers.append(_Builder.Sampler(e['Name']))

    _WriteSamplerToGmadFile(samplers,outputfile)

def _WriteSamplerToGmadFile(samplerlist, outputfile):
    f = open(outputfile, "w")
    for s in samplerlist:
        f.write(s.__repr__())
    f.close()
                           

def WrapLatticeAboutItem(maingmadfile, itemname, outputfilename):
    elementsperline = 100
    
    a = _Gmad.Lattice(maingmadfile)

    seq = a.sequence

    # remove 'lattice' from list
    try:
        seq.remove('lattice')
    except ValueError:
        pass

    # remove samplers from list
    seq = [name for name in seq if 'ampler' not in name]

    try:
        ind = seq.index(itemname)
    except ValueError:
        print('Error - ',itemname,'is not in the supplied lattice')
        return

    newseq = seq[ind:] + seq[:ind]

    # write lattice definition
    f = open(outputfilename,'w')
    timestring = '! ' + _time.strftime("%a, %d %b %Y %H:%M:%S +0000", _time.gmtime()) + '\n'
    f.write(timestring)
    f.write('! pybdsim.ModelProcessing Wrapped Lattice about '+str(itemname) +'\n')
    f.write('! LATTICE SEQUENCE DEFINITION\n\n')
    sequencechunks = _General.Chunks(newseq,elementsperline)
    linelist = []
    ti = 0
    for line in sequencechunks:
        f.write('l'+str(ti)+': line = ('+', '.join(line)+');\n')
        linelist.append('l'+str(ti))
        ti += 1
    # need to define the period before making sampler planes
    f.write('lattice: line = ('+', '.join(linelist)+');\n')
    f.write('use, lattice;\n')
    f.close()



import platform as _platform
import ROOT as _ROOT

# Check for which libraries to load
if _platform.platform().find("Darwin") != -1 :
    _ROOT.gSystem.Load("libbdsimRootEvent.dylib")
elif _platform.platform().find("Linux") != -1:
    _ROOT.gSystem.Load("libbdsimRootEvent.so")
else :
    pass
    # TODO issue warning on unknown system

from .Options   import *
from .Model     import *
from .Run       import *
from .Event     import *
from .Processed import *
from .Root      import *
# from Plots     import *


from re import split as _split

_allowedparticles = frozenset([
    'e-',
    'e+',
    'proton',
    'gamma'
])

_allowedflags = frozenset([
    '1',
    '2',
    '3'
])


class XSecBias(object):
    """
    A class for containing all information regarding cross section definitions.
    """
    def __init__(self, name, particle, processes, xsecfactors, flags):
        self.name        = self.SetName(name)
        self.particle    = self.SetParticle(particle)
        self.processlist = self.SetProcesses(processes)
        self.flaglist    = self.SetFlags(flags)
        self.xseclist    = self.SetXSecFactors(xsecfactors)
        self.CheckBiasedProcesses()

    def SetName(self, name):
        '''
        Set the bias name.
        Cannot be any upper/lowercase variant of reserved keyword "xsecBias".
        '''
        if name.lower() == 'xsecbias':
            raise ValueError("Forbidden name: "+ str(name) +".  Bias name cannot be any variant of 'xsecbias' (case insensitive)")
        return name

    def SetParticle(self, particle):
        """
        Set the particle for bias to be associated with.
        """
        if particle  in _allowedparticles:
            return particle
        else:
            raise ValueError("Unknown Particle type: " + str(particle) + ".")

    def SetProcesses(self, processes):
        '''
        Set the list of processes to be biased.
        processes hould be a space-delimited string of processes.
        '''
        processlist = _split(', |,| ', processes)
        if ('all' in processlist) and (len(processlist) > 1):
            raise ValueError("You cannot have both 'all' and another process be biased.")
        return processlist

    def SetFlags(self, flags):
        """
        Set flags. flags should be a space-delimited string of integers, 1-3,
        in the same order as the processes,
        """
        flaglist = flags.split()
        for number in flaglist:
            if number not in _allowedflags:
                raise ValueError("Unknown Flag type: " + str(number))
        return flaglist

    def SetXSecFactors(self, xsecs):
        """
        Set cross section factors.  xsecs should be a space-delimited string
        of floats, e.g. "1.0 1e13 1234.9"
        """
        xseclist = xsecs.split()
        for factor in xseclist:
            if float(factor) < 0:
                raise ValueError("Negative cross section factor: " +str(factor))
        return xseclist

    def CheckBiasedProcesses(self):
        if (len(self.xseclist) != len(self.flaglist)) or (len(self.xseclist) != len(self.processlist)):
            raise Warning("There must be a uniquely defined flag and xsecfactor to go  with every listed process.")

    def __repr__(self):
        particle = 'particle="' + self.particle + '"'
        proc     = 'proc="' + ' '.join(map(str, self.processlist)) + '"'
        xsec     = 'xsecfact={' + ','.join(map(str, self.xseclist)) + '}'
        flag     = 'flag={' + ','.join(map(str,self.flaglist)) + '}'

        s = self.name + ": xsecBias, " + ', '.join([particle,proc,xsec,flag]) + ';\n'

        return s



import numpy as _np
import optparse as _op
from collections import OrderedDict

import pysad

from .. import Builder 
from .. import Beam
from .. import Options
from .. import XSecBias

def SadFlat2GMad(inputFileName,
                 outputFileName,
                 istart=0,
                 samplers='all',
                 options = True,
                 enableSextupoles=False) :
    r = pysad.Reader.Flat(inputFileName)

    # Need nominal energy for SR calculations
    energy = 1.282
    # Emittance
    gemit    = _np.zeros(2)
    gemit[0] = 1.2e-9
    gemit[1] = 12e-12

    # Energy spread
    esprd    = 0.0006

    # particle and flip required
    particle = 'e-'
    flip     = -1.0

    # Twiss parameters
    beam = Beam.Beam(particle,energy,'gausstwiss')
    beam._SetEmittanceX(gemit[0], 'm')
    beam._SetEmittanceY(gemit[1], 'm')
    beam._SetBetaX(r.row(0)['BX'])
    beam._SetBetaY(r.row(0)['BY'])
    beam._SetAlphaX(r.row(0)['AX'])
    beam._SetAlphaY(r.row(0)['AY'])
    beam._SetSigmaE(esprd)

    # Default beam pipe radius
    defaultBpRadius = 0.05

    # create machine instance
    a = Builder.Machine(sr=False, energy0=float(energy))    

    # add the beam
    a.AddBeam(beam)


    # create options
    if options:
        o = Options.ElectronColliderOptions()
        o.SetBuildTunnel(False)
        o.SetBuildTunnelFloor(False)
        o.SetMagnetGeometryType("none")

        process = 'em'
        o.SetPhysicsList(process)
        a.AddOptions(o)

    # create name dictionary
    nameDict = {}
    unknDict = {}

    # iterate through objects and build machine
    for i in range(istart, len(r.column('NAME')), 1):

        row = r.row(i)

        # check name dictionary
        name = row['NAME']
        try:
            eCount = nameDict[name]
        except KeyError:
            nameDict[name] = 0
            eCount = nameDict[name]

        # check if name starts with a number
        prepend = ''
        if name[0].isdigit():
            prepend = 'M_'

        #if row['TYPE'] != 'QUAD' :
        #    print name,row['TYPE'],row['K1']

        if row['LENGTH'] == 0 :
            a.AddMarker(prepend + name + '_' + str(eCount))
            continue

        if row['TYPE'] == 'MARK':
            a.AddMarker(prepend + name + '_' + str(eCount))
        ###################################################################################
        elif row['TYPE'] == 'MONI':
            a.AddMarker(prepend + name + '_' + str(eCount))
        ###################################################################################
        elif row['TYPE'] == 'DRIFT':
            a.AddDrift(prepend + name + '_' + str(eCount),
                       length=row['LENGTH'],
                       aper1=defaultBpRadius)
        ###################################################################################
        elif row['TYPE'] == 'QUAD':
            if row['LENGTH'] < 1e-7:
                a.AddMarker(prepend + name + '_' + str(eCount))
            else:
                a.AddQuadrupole(prepend + name + '_' + str(eCount),
                                k1=flip*row['K1']/row['LENGTH'],
                                length=row['LENGTH'],
                                tilt=row['ROTATE'],
                                aper1=defaultBpRadius)
        ###################################################################################
        elif row['TYPE'] == 'BEND' :
            angle = row['ANGLE']
            length = row['LENGTH']
            e1    = 0
            e2    = 0
            fint  = 0
            fintx = 0
            hgap  = 0
            #print name,row['K0'],row['K1'],row['K2']
            kws = {}
            if (e1 != 0):
                kws['e1'] = e1
            if (e1 != 0):
                kws['e2'] = e2
            if (fint != 0):
                kws['fint']  = fint
            if (fintx != 0):
                kws['fintx'] = fintx
            if (hgap != 0):
                kws['hgap'] = hgap

            a.AddDipole(name,'sbend',length=row['LENGTH'],angle=angle,**kws)
        ###################################################################################
        elif row['TYPE'] == 'SEXT':
            l = row['LENGTH']
            if l < 1e-7:
                a.AddDrift(prepend + name + '_' + str(eCount),
                           length=l,
                           aper1=defaultBpRadius)
            else:
                if enableSextupoles:
                    k2in = flip*row['K2']
                else:
                    k2in = 0.0
                a.AddSextupole(prepend + name + '_' + str(eCount),
                               length=l,
                               k2=k2in,
                               aper1=defaultBpRadius)
        else:
            #    print "UNKN> ", row['TYPE']
            try :
                unknDict[row['TYPE']] += 1
            except KeyError :
                unknDict[row['TYPE']] = 1
        nameDict[name] += 1

    print(unknDict)
    a.AddSampler(samplers)
    a.Write(outputFileName)

    return a


'''
            ###################################################################################
    elif c.type[i] == 'OCTU':
        if c.data[i][c.keys['octu']['l']] > 1e-7:
            a.AddDrift(prepend + c.name[i] + '_' + str(eCount),
                       length=float(c.data[i][c.keys['octu']['l']]),
                       aper1=apertures.aper[i])
        else:
            a.AddMarker(prepend + c.name[i] + '_' + str(eCount))
            ###################################################################################
    elif c.type[i] == 'DECU':
        pass
    ###################################################################################
    elif c.type[i] == 'MULT':
        a.AddMarker(prepend + c.name[i] + '_' + str(eCount))
    ###################################################################################
    elif c.type[i] == 'HKIC':
        a.AddDrift(prepend + c.name[i] + '_' + str(eCount),
                   length=float(c.data[i][c.keys['hkic']['l']]),
                   aper1=float(apertures.aper[i]))
    ###################################################################################
    elif c.type[i] == 'VKIC':
        a.AddDrift(prepend + c.name[i] + '_' + str(eCount),
                   length=float(c.data[i][c.keys['vkic']['l']]),
                   aper1=float(apertures.aper[i]))
    ###################################################################################
    elif c.type[i] == 'SBEN':
        if c.data[i][c.keys['sben']['l']] < 1e-7:
            a.AddMarker(prepend + c.name[i] + '_' + str(eCount))
        else:
            # check for large tilt
            # if enableDipoleTiltTransform and float(c.data[i][c.keys['sben']['tilt']]) > 0.2 :
            #    print "inserting transform 3D"
            #    a.AddTransform3D(c.name[i]+"3dt_in",phi=0,theta=0,psi=float(c.data[i][c.keys['sben']['tilt']]))

            # print "SBEN> ",c.name[i],c.data[i][c.keys['sben']['tilt']]

            # check for poleface
            e1in = 0.0
            e2in = 0.0
            if enableDipolePoleFaceRotation:
                e1in = float(c.data[i][c.keys['sben']['e1']])
                e2in = float(c.data[i][c.keys['sben']['e2']])

            a.AddDipole(prepend + c.name[i] + '_' + str(eCount), 'sbend',
                        length=float(c.data[i][c.keys['sben']['l']]),
                        angle=float(c.data[i][c.keys['sben']['angle']]),
                        aper=float(apertures.aper[i]),
                        e1=e1in,
                        e2=e2in,
                        tilt=float(c.data[i][c.keys['sben']['tilt']]))


            # if enableDipoleTiltTransform and float(c.data[i][c.keys['sben']['tilt']]) > 0.2 :
            #    print "removing transform 3D"
            #    a.AddTransform3D(c.name[i]+"3dt_out",phi=0,theta=0,psi=-float(c.data[i][c.keys['sben']['tilt']]))

            ###################################################################################
    elif c.type[i] == 'LCAV':
        length = float(c.data[i][c.keys['lcav']['l']])
        deltaE = (float(c.data[i][c.keys['lcav']['E']]) - float(c.data[i - 1][c.keys['lcav']['E']])) * 1000  # MeV
        gradient = deltaE / length
        a.AddRFCavity(prepend + c.name[i] + '_' + str(eCount),
                      length=length,
                      gradient=-gradient,
                      aper1=apertures.aper[i])
    ###################################################################################
    elif c.type[i] == 'ECOL':

        if collimator == None:
            # make collimator from mad8 file
            #                print "ECOL> ",c.name[i], "mad8 file"
            print c.data[i][c.keys['ecol']['xsize']], c.data[i][c.keys['ecol']['ysize']]
            if (c.data[i][c.keys['ecol']['xsize']] != 0) and (c.data[i][c.keys['rcol']['ysize']]) != 0:
                a.AddECol(prepend + c.name[i] + '_' + str(eCount),
                          length=float(c.data[i][c.keys['ecol']['l']]),
                          xsize=float(c.data[i][c.keys['ecol']['xsize']]),
                          ysize=float(c.data[i][c.keys['ecol']['ysize']]),
                          material='Copper')
            else:
                a.AddDrift(c.name[i] + '_' + str(eCount), c.data[i][c.keys['ecol']['l']])
        else:
            # make collimator from file
            #                print "ECOL> ",c.name[i], "coll file"
            length = float(c.data[i][c.keys['rcol']['l']])
            xsize = float(collimator.getCollimator(c.name[i])['xsize'])
            ysize = float(collimator.getCollimator(c.name[i])['ysize'])
            mater = collimator.getCollimator(c.name[i])['bdsim_material']
            if xsize != 0 and ysize == 0:
                ysize = 0.2
            if xsize == 0 and ysize != 0:
                xsize = 0.2

            if (xsize != 0 and ysize != 0):
                a.AddECol(prepend + c.name[i] + '_' + str(eCount),
                          length=length,
                          xsize=xsize,
                          ysize=ysize,
                          material=mater,
                          bias=biasList)
            else:
                a.AddDrift(prepend + c.name[i] + '_' + str(eCount), float(c.data[i][c.keys['ecol']['l']]))
                ###################################################################################
    elif c.type[i] == 'RCOL':
        if collimator == None:
            #               print "RCOL> ",c.name[i], "mad8 file"
            print c.data[i][c.keys['rcol']['xsize']], c.data[i][c.keys['rcol']['ysize']]
            if (c.data[i][c.keys['rcol']['xsize']] != 0) and (c.data[i][c.keys['rcol']['ysize']]) != 0:
                a.AddRCol(prepend + c.name[i] + '_' + str(eCount),
                          length=float(c.data[i][c.keys['rcol']['l']]),
                          xsize=float(c.data[i][c.keys['rcol']['xsize']]),
                          ysize=float(c.data[i][c.keys['rcol']['ysize']]),
                          material='Copper')
            else:
                a.AddDrift(prepend + c.name[i] + '_' + str(eCount), c.data[i][c.keys['rcol']['l']])
        else:
            # make collimator from file
            #              print "RCOL> ",c.name[i], "coll file"
            length = float(c.data[i][c.keys['rcol']['l']])
            xsize = float(collimator.getCollimator(c.name[i])['xsize'])
            ysize = float(collimator.getCollimator(c.name[i])['ysize'])
            mater = collimator.getCollimator(c.name[i])['bdsim_material']
            if xsize != 0 and ysize == 0:
                ysize = 0.2
            if xsize == 0 and ysize != 0:
                xsize = 0.2

            if (xsize != 0 and ysize != 0):
                a.AddRCol(prepend + c.name[i] + '_' + str(eCount),
                          length=length,
                          xsize=xsize,
                          ysize=ysize,
                          material=mater,
                          bias=biasList)
            else:
                a.AddDrift(prepend + c.name[i] + '_' + str(eCount), float(c.data[i][c.keys['rcol']['l']]))
                ###################################################################################
'''


from scipy import constants as _con
import numpy as _np
import os.path as _path
import time as _time

import pybdsim.Data as _Data

def BdsimPrimaries2Ptc(inputfile, outfile=None, start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run an an input and creates
    a PTC inrays file from the primary particle tree. If no output file
    name is given, it will be <inputfile>.madx (stripping off the .root).
    
    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate (-1 means all)
    :type  ninrays: int
    """
    if outfile is None:
        outfile = inputfile.rstrip(".root")
        outfile += ".madx"

    BdsimSampler2Ptc(inputfile, outfile, "Primary", start, ninrays)

def BdsimSampler2Ptc(inputfile, outfile, samplername, start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run an an input and creates
    a PTC inrays file from the sampler particle tree. Converts primary particles only.
    
    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param samplername: sampler name in BDSIM root file
    :type  samplername: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate
    :type  ninrays: int
    """
    if not (outfile[-5:] == ".madx"):
        outfile = outfile + ".madx"

    # specify if primaries as there's a minor difference in conversion. The primary sampler can use the nominal
    # energy from the beam tree, however a specified sampler may be after an energy changing element (e.g
    # degrader or RF cavity), so the energy is taken to be the mean particle energy in that sampler. Note that
    # this is susceptible to incorrect conversion at low statistics.
    if samplername == "Primary":
        sampler_coords = _LoadBdsimCoordsAndConvert(inputfile, samplername, start, ninrays, isPrimaries=True)
    else:
        sampler_coords = _LoadBdsimCoordsAndConvert(inputfile, samplername, start, ninrays, isPrimaries=False)

    outfile = open(outfile, 'w')

    nentries = len(sampler_coords[0])
    headstr = "! PTC format inrays file of " + str(nentries)
    headstr += " initial coordinates generated from primaries in BDSIM sampler " + samplername + " on " + _time.strftime("%c") + "\n"

    outfile.writelines(headstr)
    for n in range(0, nentries):  # n denotes a given particle
        s = 'ptc_start'
        s += ', x=' + repr(sampler_coords[0][n])
        s += ', px=' + repr(sampler_coords[1][n])
        s += ', y=' + repr(sampler_coords[2][n])
        s += ', py=' + repr(sampler_coords[3][n])
        s += ', t=' + repr(sampler_coords[4][n])
        s += ', pt=' + repr(sampler_coords[5][n])
        s += ';\n'
        outfile.writelines(s)

    outfile.close()

def BdsimPrimaries2BdsimUserFile(inputfile, outfile, start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run and creates
    a BDSIM userFile file from the primary particle tree.
    
    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate
    :type  ninrays: int
    
    Writes the following columns to file:
      x[m] xp[rad] y[m] yp[rad] t[ns] E[GeV]
    E is the total particle energy.
    """
    BdsimSampler2BdsimUserFile(inputfile, outfile, "Primary", start, ninrays)

def BdsimSampler2BdsimUserFile(inputfile, outfile, samplername, start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run and creates
    a BDSIM userFile file from the sampler particle tree.
    
    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param samplername: sampler name in BDSIM root file
    :type  samplername: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate
    :type  ninrays: int
    
    Writes the following columns to file:
      x[m] xp[rad] y[m] yp[rad] t[ns] E[GeV]
    
    E is the total particle energy. The t column is the time in the given 
    sampler minus the mean time for that sampler. If not mean subtracted, 
    the particles may be significantly offset from the primary position.
    """
    if not (outfile[-4:] == ".dat"):
        outfile = outfile + ".dat"

    if not _path.isfile(inputfile):
        raise IOError("file \"{}\" not found!".format(inputfile))
    else:
        print("Loading input file: ", inputfile)
        data = _Data.Load(inputfile)

    x,xp,y,yp,tof,E,pid = _ExtractSamplerCoords(data, samplername)

    # subtract mean time as non-primary sampler will be at a finite T in the lattice - should be centred around 0.
    if samplername != "Primary":
        meanT = _np.mean(tof)
        tof = tof - meanT
    nentries = len(x)

    x,xp,y,yp,t,E = _TruncateCoordinates(x,xp,y,yp,tof,E,ninrays,start)

    outfile = open(outfile, 'w')
    for n in range(0, nentries):  # n denotes a given particle
        s =  ' ' + repr(x[n])
        s += ' ' + repr(xp[n])
        s += ' ' + repr(y[n])
        s += ' ' + repr(yp[n])
        s += ' ' + repr(tof[n])
        s += ' ' + repr(E[n])
        s += '\n'
        outfile.writelines(s)

    outfile.close()

def BdsimPrimaries2Madx(inputfile,outfile,start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run an an input and creates
    a MADX inrays file from the primary particle tree.

    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate, default is all available
    :type  ninrays: int
    """
    if not (outfile[-5:] == ".madx"):
        outfile = outfile+".madx"

    primary_coords = _LoadBdsimCoordsAndConvert(inputfile, "Primary", start, ninrays, isPrimaries=True)

    outfile = open(outfile,'w' )

    nentries =  len(primary_coords[0])
    headstr  = "! MadX format inrays file of "+str(nentries)
    headstr += " initial coordinates generated from BDSIM output on "+_time.strftime("%c")+"\n"

    outfile.writelines(headstr)
    for n in range(0,nentries):               # n denotes a given particle
        s  =  'start'
        s += ', x='  + repr(primary_coords[0][n])
        s += ', px=' + repr(primary_coords[1][n])
        s += ', y='  + repr(primary_coords[2][n])
        s += ', py=' + repr(primary_coords[3][n])
        s += ', t='  + repr(primary_coords[4][n])
        s += ', pt=' + repr(primary_coords[5][n])
        s += ';\n'
        outfile.writelines(s)

    outfile.close()

def BdsimPrimaries2Mad8(inputfile,outfile,start=0, ninrays=-1):
    """
    Takes .root file generated from a BDSIM run an an input and creates
    a MAD8 inrays file from the primary particle tree.

    :param inputfile: root format output from BDSIM run
    :type  inputfile: str
    :param outfile: filename for the inrays file
    :type  outfile: str
    :param start: starting primary particle index
    :type  start: int
    :param ninrays: total number of inrays to generate
    :type  ninrays: int
    """
    if not (outfile[-5:] == ".mad8"):
        outfile = outfile+".mad8"

    primary_coords = _LoadBdsimCoordsAndConvert(inputfile, "Primary", start, ninrays, isPrimaries=True)

    outfile = open(outfile,'w' )

    nentries =  len(primary_coords[0])
    headstr  = "! Mad8 format inrays file of "+repr(nentries)
    headstr += " initial coordinates generated from BDSIM output on "+_time.strftime("%c")+"\n"

    outfile.writelines(headstr)
    for n in range(0,nentries):    #n denotes a given particle
        s  =  'START'
        s += ', X='  + repr(primary_coords[0][n])
        s += ', PX=' + repr(primary_coords[1][n])
        s += ', Y='  + repr(primary_coords[2][n])
        s += ', &\n'                             #line continuation needed to obey FORTRAN 80 char input limit
        s += 'PY=' + repr(primary_coords[3][n])
        s += ', T='  + repr(primary_coords[4][n])
        s += ', DELTAP=' + repr(primary_coords[5][n])
        s += '\n'
        outfile.writelines(s)

    outfile.close()

def _LoadBdsimCoordsAndConvert(inputfile, samplername, start, ninrays, isPrimaries):
    """
    Load BDSIM coordinates and convert to PTC format.
    """
    if not _path.isfile(inputfile):
        raise IOError("file \"{}\" not found!".format(inputfile))
    else:
        print("Loading input file: ", inputfile)
        data = _Data.Load(inputfile)

    #Get sampler/primaries data
    x,xp,y,yp,tof,E,pid = _ExtractSamplerCoords(data, samplername)

    #Get particle pdg number
    pid  =  int(_np.mean(pid))  #cast to int to match pdg id

    #Particle mass needed for calculating momentum, in turn needed for dE.
    mass = 0
    if pid == 2212:                                     #proton
        mass = _con.proton_mass * _con.c**2 / _con.e / 1e9
    elif (pid == 11) or (pid == -11):                   #electron / positron
        mass = _con.electron_mass * _con.c**2 / _con.e / 1e9
    elif (pid == 13) or (pid == -13):                   #mu- / mu+
        mass = 0.1056583745
    elif (pid == 1000020040):                           # helium4 nuclei
        mass = 3.756671525    # value calculated by Keenan Ngo 02/2020
    elif (pid == 1000060120):                           # carbon12 nuclei
        mass = 11.27001475    # value calculated by Keenan Ngo 02/2020

    #TODO: Add more particle masses and particle numbers as needed.
    # Note PDGID scheme for ions is 10LZZZAAAI - L is no. strange quarks (0 realistically), Z is no. protons,
    # A is no. neutrons+protons, and I is excited state, 0 being ground.

    if mass == 0:
        raise ValueError('Unknown particle species.')

    if isPrimaries:
        # use design energy for primaries as a significant mean offset can exist with small number of particles
        beam = _Data.BeamData(data)
        Em = beam.beamEnergy
    else:
        # Use the mean energy as there may have been a designed energy change (from RF, degrader, etc)
        Em = _np.mean(E)

    beta = _np.sqrt(1 - (mass/Em)**2)

    p           = _np.sqrt(E**2 - _np.full_like(E, mass)**2)
    p0          = _np.sqrt(Em**2 - mass**2)
    tofm        = _np.mean(tof)

    # BDSIM output momenta are normalised w.r.t the individual particle
    # momentum.  But PTC expects normalised w.r.t the reference momentum.
    xp *= (p / p0)
    yp *= (p / p0)

    #Use deltap and pathlength as needed for the time=false flag in PTC
    #Reference on p.201 of the MADX User's Reference Manual V5.03.07
    dE          = (p-p0)/p0
    t           = beta*(tof-_np.full(len(tof),tofm))*1.e-9*_con.c    #c is sof and the 1.e-9 factor is nm to m conversion

    x,xp,y,yp,t,dE = _TruncateCoordinates(x,xp,y,yp,t,dE,ninrays,start)

    #Agglomerate the coordinate arrays and return resulting superarray
    coords = _np.stack([x,xp,y,yp,t,dE])
    return coords

def _ExtractSamplerCoords(data, samplername):
    """
    Extract sampler coordinates.
    """
    if samplername != "Primary":
        # add . to the sampler name to match branch names from file
        if samplername[-1] != ".":
            samplername += "."
        # check branch exists
        allSamplers = data.GetSamplerNames()
        if not samplername in allSamplers:
            msg = "Sampler " + samplername + " not found in inputfile."
            raise ValueError(msg)

    sampler = _Data.SamplerData(data, samplername)

    # get particle coords and filter out secondaries
    primary = sampler.data['parentID'] == 0
    x   = sampler.data['x'][primary]
    xp  = sampler.data['xp'][primary]
    y   = sampler.data['y'][primary]
    yp  = sampler.data['yp'][primary]
    tof = sampler.data['T'][primary]
    E   = sampler.data['energy'][primary]
    pid = sampler.data['partID'][primary]

    return x,xp,y,yp,tof,E,pid

def _TruncateCoordinates(x,xp,y,yp,t,dE,ninrays,start):
    # Truncate the arrays to the desired length
    if ninrays < 0:
        x = x[start:]
        y = y[start:]
        xp = xp[start:]
        yp = yp[start:]
        t = t[start:]
        dE = dE[start:]

    else:
        if start > 0:
            ninrays += start
        x = x[start:ninrays]
        y = y[start:ninrays]
        xp = xp[start:ninrays]
        yp = yp[start:ninrays]
        t = t[start:ninrays]
        dE = dE[start:ninrays]

    return x,xp,y,yp,t,dE



"""
Converts ROOT histogram to matplotlib figure
"""

import matplotlib.pyplot as _plt
import matplotlib.cm as _cm
import numpy as _np
import matplotlib.ticker as _ticker
from matplotlib.colors import LogNorm
from mpl_toolkits.axes_grid1 import make_axes_locatable



'''
Wrapper classes for ROOT data to be plotted properly in matplotlib
'''

class TTree :
    def __init__(self, r2n_tree):
        '''
        :param r2n_tree: root_numpy converted tree
        '''
        self.data = r2n_tree

    def Plot(self, keyX, keyY, keyYErr = '', opt = '', label = ''):
        if keyYErr == '' :
            _plt.plot(self.data[keyX], self.data[keyY], opt, label = label)
        else :
            _plt.errorbar(self.data[keyX], self.data[keyY], self.data[keyYErr], opt, label = label)

    def Keys(self):
        return self.data.dtype.names

class TH1 :
    # TODO : TH1 Deal with under/overflows properly
    def __init__(self, hist):
        self.hist   = hist

        # extract meta data
        self.name   = hist.GetName()
        self.title  = hist.GetTitle()
        self.labelX = hist.GetXaxis().GetTitle()
        self.labelY = hist.GetYaxis().GetTitle()

        #Extract statistics box data
        self.mean  = hist.GetMean()
        self.stdev = hist.GetStdDev()

        # extract data
        nbinsx   = hist.GetNbinsX()
        self.entries   = hist.GetEntries()
        self.widths   = _np.zeros(nbinsx)
        self.centres  = _np.zeros(nbinsx)
        self.lowedge  = _np.zeros(nbinsx)
        self.highedge = _np.zeros(nbinsx)
        self.contents = _np.zeros(nbinsx)
        self.errors   = _np.zeros(nbinsx)

        for i in range(nbinsx):
            self.widths[i]   = hist.GetXaxis().GetBinWidth(i)
            self.lowedge[i]  = hist.GetBinLowEdge(i+1)
            self.highedge[i] = hist.GetBinLowEdge(i+1)
            self.centres[i]  = hist.GetBinCenter(i+1)
            self.contents[i] = hist.GetBinContent(i+1)
            self.errors[i]   = hist.GetBinError(i+1)

    def Plot(self,opt ='hist', logx=False, logy=False, outputfilename=None, axlabels=None, color='b'):
        '''
        ROOT like plotting options for convenience
        :param opt: hist|e1
        :return:
        '''
        _plt.figure(num=1, figsize=(11,7), facecolor="w", edgecolor="w")

        if opt.find('hist') != -1 :
            self.PlotHist()
        elif opt.find('line') != -1 :
            self.PlotPlot()
        elif opt.find('e1') != -1:
            self.PlotErrorbar(color)

        self._SetLabels(axlabels)
        self._StatBox()

        if logx or logy:
            _LogAxes(logx, logy)

        if outputfilename != None:
            _SaveFigure(outputfilename)

    def PlotPlot(self):
        _plt.plot(self.centres,self.contents)

    def PlotErrorbar(self, edgecolor='none', color='b', label='', colorin='b'):
        p = _plt.plot(self.centres+self.widths/2, self.contents, ls="steps", color=colorin)
        _plt.errorbar(self.centres,self.contents, self.errors, fmt="", color=colorin, label=label)

    def PlotBar(self, edgecolor='none', color='b', label=''):
        _plt.bar(self.lowedge, self.contents, self.widths, edgecolor=edgecolor, color=color, label=label)

    def PlotHist(self, edgecolor='none', color='b', label=''):
        _plt.hist(self.centres, self.lowedge, weights=self.contents, edgecolor=edgecolor, color=color, label=label)

    def _SetLabels(self, axlabels):
        if axlabels:
            _plt.xlabel(axlabels[0])
            _plt.ylabel(axlabels[1])
        else:
            _plt.xlabel(self.labelX)
            _plt.ylabel(self.labelY)

    def _StatBox(self):
        _plt.plot([],[], linestyle="None", label=self.title)
        _plt.plot([],[], linestyle="None", label=r"Entries".ljust(15)+_LegNum(self.entries))
        _plt.plot([],[], linestyle="None", label=r"Mean".ljust(15)+_LegNum(self.mean))
        _plt.plot([],[], linestyle="None", label=r"Std Dev".ljust(14)+_LegNum(self.stdev))
        leg = _plt.legend(loc=1, frameon=True, edgecolor="k",
                    handlelength=0.05, fontsize="xx-small")
        if leg:
            leg.draggable(True,update='bbox')



class TH2 :
    # TODO : TH2 Deal with under/overflows properly
    def __init__(self,hist):
        self.hist   = hist

        # extract meta data
        self.name   = hist.GetName()
        self.title  = hist.GetTitle()
        self.labelX = hist.GetXaxis().GetTitle()
        self.labelY = hist.GetYaxis().GetTitle()

        #Extract statistics box data
        self.meanx  = hist.GetMean(1)
        self.stdevx = hist.GetStdDev(1)
        self.meany  = hist.GetMean(2)
        self.stdevy = hist.GetStdDev(2)

        # extract data
        nbinsx   = hist.GetNbinsX()
        nbinsy   = hist.GetNbinsY()
        self.entries   = hist.GetEntries()
        self.xwidths   = _np.zeros(nbinsx)
        self.xcentres  = _np.zeros(nbinsx)
        self.xlowedge  = _np.zeros(nbinsx)
        self.xhighedge = _np.zeros(nbinsx)
        self.ywidths   = _np.zeros(nbinsy)
        self.ycentres  = _np.zeros(nbinsy)
        self.ylowedge  = _np.zeros(nbinsy)
        self.yhighedge = _np.zeros(nbinsy)
        self.contents = _np.zeros((nbinsx,nbinsy))
        self.errors   = _np.zeros((nbinsx,nbinsy))

        for i in range(nbinsx):
            self.xwidths[i]   = hist.GetXaxis().GetBinWidth(i+1)
            self.xlowedge[i]  = hist.GetXaxis().GetBinLowEdge(i+1)
            self.xhighedge[i] = hist.GetXaxis().GetBinLowEdge(i+2)
            self.xcentres[i]  = hist.GetXaxis().GetBinCenter(i+1)

        for i in range(nbinsy):
            self.ywidths[i]   = hist.GetYaxis().GetBinWidth(i+1)
            self.ylowedge[i]  = hist.GetYaxis().GetBinLowEdge(i+1)
            self.yhighedge[i] = hist.GetYaxis().GetBinLowEdge(i+2)
            self.ycentres[i]  = hist.GetYaxis().GetBinCenter(i+1)

        for i in range(nbinsx) :
            for j in range(nbinsy) :
                self.contents[i,j] = hist.GetBinContent(i+1,j+1)
                self.errors[i,j]   = hist.GetBinError(i+1,j+1)

    def Plot(self, logx=False, logy=False, logz=False, cmap="jet", axlabels=None, outputfilename=None):
        _plt.figure(num=1, figsize=(12,6), facecolor="w", edgecolor="w")

        self.PlotColz(logx,logy, logz, cmap, axlabels) #In the future there may be other plotting options

        self._SetLabels(axlabels) #TODO: Make the label setting more robust
        self._StatBox()

        if outputfilename != None:
            _SaveFigure(outputfilename)

    def PlotColz(self, logx=False, logy=False, logz=False, cmap="jet", axlabels=None):
        _plt.figure(num=1, figsize=(12,6), facecolor="w", edgecolor="w")
        ax = _plt.gca()
        if axlabels:
            _plt.xlabel(axlabels[0])
            _plt.ylabel(axlabels[1])

        xx, yy = _np.meshgrid(self.xcentres,self.ycentres)
        ccmap = _cm.get_cmap(cmap, 20)  #Get 20 discrete colors from the colormap - looks nicer
        cts = self.contents.transpose() #shortcut
        minVal = _np.min(cts[_np.nonzero(cts)])

        if logz:
            _plt.pcolormesh(xx,yy,cts, cmap=ccmap, norm=LogNorm(vmin=minVal, vmax=cts.max()))
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = _plt.colorbar(cax=cax,format=_ticker.FuncFormatter(_fmtCbar),ticks=_ticker.LogLocator())
        else:
            _plt.pcolormesh(xx,yy,cts)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = _plt.colorbar(cax=cax, format=_ticker.FuncFormatter(_fmtCbar))

        _plt.axes(ax) #Make the plot axes active again

        if logx:
            ax.set_xscale("log", nonposy="clip", nonposx="clip")
        if logy:
            ax.set_yscale("log", nonposy="clip", nonposx="clip")

        cb.ax.tick_params(labelsize="xx-small")

    def _SetLabels(self, axlabels):
        if axlabels:
            _plt.xlabel(axlabels[0])
            _plt.ylabel(axlabels[1])
        else:
            _plt.xlabel(self.labelX)
            _plt.ylabel(self.labelY)

    def _StatBox(self):
        _plt.plot([],[], linestyle="None", label=self.title)
        _plt.plot([],[], linestyle="None", label=r"Entries".ljust(15)+_LegNum(self.entries))
        _plt.plot([],[], linestyle="None", label=r"Mean x".ljust(15)+_LegNum(self.meanx))
        _plt.plot([],[], linestyle="None", label=r"Mean y".ljust(15)+_LegNum(self.meany))
        _plt.plot([],[], linestyle="None", label=r"Std Dev x".ljust(14)+_LegNum(self.stdevx))
        _plt.plot([],[], linestyle="None", label=r"Std Dev y".ljust(14)+_LegNum(self.stdevy))
        leg = _plt.legend(bbox_to_anchor=(0.68, 0.62), bbox_transform=_plt.gcf().transFigure,
                          loc="lower left", frameon=True, edgecolor="k", handlelength=0.05,
                          fontsize="xx-small", framealpha=1)
        #if leg:
        #    leg.draggable(True, update='loc')
        #leg.set_zorder(20)

#Unbound utility functions
def _LogAxes(x=False, y=False):
    ax = _plt.gca()
    if x:
        ax.set_xscale("log", nonposy='clip')
    if y:
        ax.set_yscale("log", nonposy='clip')

def _LegNum(number): #TODO: make more elegant with string formatting
    n = number #shortcut
    l = str(n)

    if "e" in l:              #Numbers that are already in scientific notation
        ln  = l.split("e")[0]
        le  = l.split("e")[1]
        try:
            lnr = ln[:7]
        except:
            lnr = ln

        nas = lnr+r"x10$^{"+le+r"}$"

    elif n > 1.e6:             #Large numbers (eg. 2375)
        if l[:-1][-1]==("."):  #Take into account trailing .0s (eg. 2375.0)
            ll = len(l)-2
        else:
            ll = len(l)        #ll is the power of 10 (e.g 3)
        lf = l[0]              #The first number will be the integer base (eg. 2)
        try:
            ls = l[1:6]        #Get the fractional base (e.g 375)
        except:
            ls = l[1:]
        if not ls:             #Or add a zero if no fractional base
            ls = "0"
        nas = lf+r"."+ls+r"x10$^{"+str(ll-1)+r"}$" #Put a decimal point and a power of 10

    elif n < 1.e-6:               #Small numbers (eg. 0.0000456)
        lsp  = l.split(".")[1]    #Get digits after the decimal point
        lr   = lsp.split("0")[-1]         #Get the part with no leading zeros (e.g 456)
        nz   = len(lsp.split("0")[-1])-1  #Count the leading zeros
        lf   = lr[0]                      #Get the desired integer base

        try:
            ls   = lr[1:6]     #Geth the fractional base
        except:
            ls   = lr[1:]
        if not ls:
            ls = "0"           #If no fractional base add a zero
        nas  = lf+r"."+ls+r"x10$^{-"+str(nz+2)+r"}$" #Put decimal point and neg power of 10

    else:
        try:
            nas = l[:7]        #If number is small/big enough print without alteration
        except:
            nas = l

    return nas.ljust(11)


def _fmtCbar(x, pos): #Format in scientific notation and make vals < 1 = 0
    if float(x) == 1.0:
        fst = r"$1$" #For a histogram valuesa smalled that 1 are set to 0
    else:            #Such values are set as dummies to allow log plots
        a, b = '{:.0e}'.format(x).split('e')
        b = int(b)
        fst = r'$10^{{{}}}$'.format(b)
    return fst

def _SaveFigure(outputfilename):
    if '.' in outputfilename:
        outputfilename = outputfilename.split('.')[0]
    _plt.savefig(outputfilename+'.pdf')
    #_plt.savefig(outputfilename+'.png')




import pickle as _pkl
import pylab  as _pl
import pymad8 as _pymad8
import pybdsim as _pybdsim
import matplotlib.pyplot as _plt
import numpy as _np
from os.path import isfile as _isfile
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages
import datetime as _datetime


# Predefined dicts for making the standard plots,
# format = (mad8_optical_var_name, bdsim_optical_var_name, bdsim_optical_var_error_name, legend_name)

_BETA =    {"bdsimdata"  : ("Beta_x", "Beta_y"),
            "bdsimerror" : ("Sigma_Beta_x","Sigma_Beta_y"),
            "mad8"       : ("betx", "bety"),
            "legend"     : (r'$\beta_{x}$', r'$\beta_{y}$'),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\beta_{x,y}$ / m",
            "title"      : "Beta"
            }

_ALPHA =   {"bdsimdata"  : ("Alpha_x", "Alpha_y"),
            "bdsimerror" : ("Sigma_Alpha_x","Sigma_Alpha_y"),
            "mad8"       : ("alfx", "alfy"),
            "legend"     : (r'$\alpha_{x}$', r'$\alpha_{y}$'),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\alpha_{x,y}$ / m",
            "title"      : "Alpha"
           }

_DISP  =   {"bdsimdata"  : ("Disp_x", "Disp_y"),
            "bdsimerror" : ("Sigma_Disp_x","Sigma_Disp_y"),
            "mad8"       : ("dx", "dy"),
            "legend"     : (r"$\eta_{x}$", r"$\eta_{y}$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\eta_{x,y} / m$",
            "title"      : "Dispersion"
            }

_DISP_P=   {"bdsimdata"  : ("Disp_xp", "Disp_yp"),
            "bdsimerror" : ("Sigma_Disp_xp","Sigma_Disp_yp"),
            "mad8"       : ("dpx", "dpy"),
            "legend"     : (r"$\eta_{p_x}$", r"$\eta_{p_x}$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\eta_{p_{x},p_{y}}$ / m",
            "title"      : "Momentum_Dispersion"
            }

_SIGMA =   {"bdsimdata"  : ("Sigma_x", "Sigma_y"),
            "bdsimerror" : ("Sigma_Sigma_x","Sigma_Sigma_y"),
            "mad8"       : ("", ""),
            "legend"     : (r"$\sigma_{x}$",r"$\sigma_{y}$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\sigma_{x,y}$ / m",
            "title"      : "Sigma"
            }

_SIGMA_P = {"bdsimdata"  : ("Sigma_xp", "Sigma_yp"),
            "bdsimerror" : ("Sigma_Sigma_xp","Sigma_Sigma_yp"),
            "mad8"       : ("", ""),
            "legend"     : (r"$\sigma_{xp}$",r"$\sigma_{yp}$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\sigma_{xp,yp}$ / rad",
            "title"      : "SigmaP"
            }

_MEAN    = {"bdsimdata"  : ("Mean_x", "Mean_y"),
            "bdsimerror" : ("Sigma_Mean_x","Sigma_Mean_y"),
            "mad8"       : ("x", "y"),
            "legend"     : (r"$\overline{x}$", r"$\overline{y}$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\bar{x,y}$ / m",
            "title"      : "Mean"
            }

_EMITT   = {"bdsimdata"  : ("Emitt_x", "Emitt_y"),
            "bdsimerror" : ("Sigma_Emitt_x","Sigma_Emitt_y"),
            "mad8"       : ("", ""),
            "legend"     : (r"$\epsilon_x$", r"$\epsilon_y$"),
            "xlabel"     : "S / m",
            "ylabel"     : r"$\epsilon_{x,y}$ / m",
            "title"      : "Emittance"
            }

# use closure to avoid tonnes of boilerplate code as happened with
# MadxBdsimComparison.py
def _make_plotter(plot_info_dict):
    def f_out(mad8opt, bdsopt, functions=None, postfunctions=None, survey=None, figsize=(9,5), xlim=(0,0), **kwargs):

        # Get the initial N for the bdsim
        N = str(int(bdsopt['Npart'][0]))  # number of primaries.

        # labels for plot legends
        mad8legendx = r'MAD8 ' + plot_info_dict['legend'][0]
        mad8legendy = r'MAD8 ' + plot_info_dict['legend'][1]
        bdslegendx  = r'BDSIM ' + plot_info_dict['legend'][0] + ' ; N = ' + N
        bdslegendy  = r'BDSIM ' + plot_info_dict['legend'][1] + ' ; N = ' + N

        # mad8 data from correct source
        if plot_info_dict["title"] == "Sigma":
            sigmaX,sigmaY,sigmaXP,sigmaYP = _CalculateSigmas(mad8opt)
            mad8Xdata = sigmaX
            mad8Ydata = sigmaY

            mad8Xdata = _np.sqrt(mad8opt['envel'].getColumn('s11'))
            mad8Ydata = _np.sqrt(mad8opt['envel'].getColumn('s33'))

            mad8s     = mad8opt['envel'].getColumn('suml')
            mad8legendx += '(calculated)'
            mad8legendy += '(calculated)'
        elif plot_info_dict["title"] == "SigmaP":
            sigmaX,sigmaY,sigmaXP,sigmaYP = _CalculateSigmas(mad8opt)
            mad8Xdata = sigmaXP
            mad8Ydata = sigmaYP

            mad8Xdata = _np.sqrt(mad8opt['envel'].getColumn('s22'))
            mad8Ydata = _np.sqrt(mad8opt['envel'].getColumn('s44'))

            mad8s     = mad8opt['envel'].getColumn('suml')
            mad8legendx += '(calculated)'
            mad8legendy += '(calculated)'

        elif plot_info_dict["title"] == "Emittance":
            emitX, emitY = _CalculateEmittance(mad8opt)
            mad8Xdata = emitX
            mad8Ydata = emitY
            mad8s     = mad8opt['twiss'].getColumn('suml')
        else:
            mad8Xdata = mad8opt['twiss'].getColumn(plot_info_dict['mad8'][0])
            mad8Ydata = mad8opt['twiss'].getColumn(plot_info_dict['mad8'][1])
            mad8s     = mad8opt['twiss'].getColumn('suml')

        # the figure
        plot = _plt.figure(plot_info_dict["title"], figsize=figsize, **kwargs)

        # mad8 plot
        _plt.plot(mad8s, mad8Xdata, 'b--', label=mad8legendx)
        _plt.plot(mad8s, mad8Ydata, 'g--', label=mad8legendy)

        # bds plot
        _plt.errorbar(bdsopt['S'],
                      bdsopt[plot_info_dict['bdsimdata'][0]],
                      yerr=bdsopt[plot_info_dict['bdsimerror'][0]],
                      label=bdslegendx,
                      capsize=3, ls='', marker='x', color='b', **kwargs)

        _plt.errorbar(bdsopt['S'],
                      bdsopt[plot_info_dict['bdsimdata'][1]],
                      yerr=bdsopt[plot_info_dict['bdsimerror'][1]],
                      label=bdslegendy,
                      capsize=3, ls='', marker='x', color='g', **kwargs)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(plot_info_dict['ylabel'])
        axes.set_xlabel(plot_info_dict['xlabel'])
        axes.legend(loc='best')
        axes.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

        if survey is None:
            survey = mad8opt
        _CallUserFigureFunctions(functions)
        _AddSurvey(plot, survey)
        _CallUserFigureFunctions(postfunctions)
        if xlim != (0, 0):
            _plt.xlim(xlim)

        plot.sca(plot.axes[0])
        _plt.show(block=False)

        #if survey is not None:
        #    _pymad8.Plot.AddMachineLatticeToFigure(plot, survey)

        _plt.show(block=False)
        return plot
    return f_out


PlotBeta   = _make_plotter(_BETA)
PlotAlpha  = _make_plotter(_ALPHA)
PlotDisp   = _make_plotter(_DISP)
PlotDispP  = _make_plotter(_DISP_P)
PlotSigma  = _make_plotter(_SIGMA)
PlotSigmaP = _make_plotter(_SIGMA_P)
PlotMean   = _make_plotter(_MEAN)
PlotEmitt  = _make_plotter(_EMITT)


def _CalculateSigmas(mad8opt):
    rgamma, rbeta, emitXN0, emitYN0 = _CalculateNEmittance(mad8opt)

    E = mad8opt['comm'].getColumn('E')
    E0 = E[0]
    sigE = mad8opt['beam']['esprd']

    sige = sigE*E0/E # absolute energy spread is constant, fractional decreases (TODO need the energy spread from MAD8)

    sigmaX = _np.sqrt(emitXN0*mad8opt['twiss'].getColumn('betx')/(rbeta*rgamma)+mad8opt['twiss'].getColumn('dx')**2*sige**2)
    sigmaY = _np.sqrt(emitYN0*mad8opt['twiss'].getColumn('bety')/(rbeta*rgamma)+mad8opt['twiss'].getColumn('dy')**2*sige**2)

    sigmaXP = _np.sqrt(emitXN0 / (rbeta * rgamma) * (mad8opt['twiss'].getColumn('alfx') ** (2) + 1) / mad8opt['twiss'].getColumn(
            'betx') + (mad8opt['twiss'].getColumn('dpx') ** (2)) * (sige ** (2)))
    sigmaYP = _np.sqrt(emitYN0 / (rbeta * rgamma) * (mad8opt['twiss'].getColumn('alfy') ** (2) + 1) / mad8opt['twiss'].getColumn(
            'bety') + (mad8opt['twiss'].getColumn('dpy') ** (2)) * (sige ** (2)))
    return sigmaX, sigmaY, sigmaXP, sigmaYP

def _CalculateNEmittance(mad8opt):
    # Own calculation of beam sizes
    emitX0 = mad8opt['beam']['ex']
    emitY0 = mad8opt['beam']['ey']
    particle =  mad8opt['beam']['particle']
    if particle == 'electron' or particle == 'positron':
        mass = 0.5109989461
    elif particle == 'proton':
        mass = 938.2720813
    else:  # default is mad8 default particle mass.
        mass = 0.5109989461

    e = mad8opt['comm'].getColumn('E')
    rgamma = e / (mass / 1e3) 
    rbeta = _np.sqrt(1 - 1.0 / rgamma ** 2)

    emitXN0 = emitX0 * rgamma[0] * rbeta[0]
    emitYN0 = emitY0 * rgamma[0] * rbeta[0]
    return rgamma, rbeta, emitXN0, emitYN0

def _CalculateEmittance(mad8opt):
    rgamma, rbeta, emitXN0, emitYN0 = _CalculateNEmittance(mad8opt)
    emitX = emitXN0 / (rbeta * rgamma)
    emitY = emitYN0 / (rbeta * rgamma)
    return emitX, emitY

def _CheckFilesExist(twiss, envel, bdsim):
    '''
    Otherwise such errors are too cryptic.
    '''
    if not _isfile(twiss):
        raise IOError("File not found: ", twiss)
    if not _isfile(envel):
        raise IOError("File not found: ", envel);
    if isinstance(bdsim, str) and not _isfile(bdsim):
        raise IOError("File not found: ", bdsim)

def _GetBDSIMOptics(optics):
    '''
    Takes a BDSAscii instance.
    Return a dictionary of lists matching the variable with the list of values.
    '''

    optvars = {}
    for variable in optics.names:
        datum = getattr(optics, variable)()
        optvars[variable] = datum
    return optvars


def Mad8VsBDSIM(twiss, envel, bdsim, survey=None, functions=None,
                postfunctions=None, figsize=(10, 5), xlim=(0,0),
                saveAll=True, outputFileName=None,
                particle="electron", energySpread=1e-4, ex=1e-8, ey=1e-8):
    """
    Compares Mad8 and BDSIM optics variables.

    +-----------------+---------------------------------------------------------+
    | **Parameters**  | **Description**                                         |
    +-----------------+---------------------------------------------------------+
    | twiss           | Mad8 twiss file                                         |
    +-----------------+---------------------------------------------------------+
    | bdsim           | Optics root file (from rebdsimOptics or rebdsim).       |
    +-----------------+---------------------------------------------------------+
    | functions       | Hook for users to add their functions that are called   |
    |                 | immediately prior to the addition of the plot. Use a    |
    |                 | lambda function to add functions with arguments. Can    |
    |                 | be a function or a list of functions.                   |
    +-----------------+---------------------------------------------------------+
    | figsize         | Figure size for all figures - default is (12,5)         |
    +-----------------+---------------------------------------------------------+
    | xlim            | Set xlimit for all figures                              |
    +-----------------+---------------------------------------------------------+
    | particle        | Beam particle type to determine particle mass, required |
    |                 | for beam size calculation - default is electron.        |
    +-----------------+---------------------------------------------------------+
    | energySpread    | Energy spread used in beam size calculation - default   |
    |                 | is 1e-4.                                                |
    +-----------------+---------------------------------------------------------+
    | ex / ey         | Horizontal / vertical emittance used in beam size       |
    |                 | calculation - default is 1e-8.                          |
    +-----------------+---------------------------------------------------------+
    """

    _CheckFilesExist(twiss, envel, bdsim)

    fname = _pybdsim._General.GetFileName(bdsim) # cache file name
    if fname == "":
        fname = "optics_report"
    
    # load mad8 optics
    mad8reader   = _pymad8.Output.OutputReader() 
    [com, twissL] = mad8reader.readFile(twiss,'twiss')
    [com, envelL] = mad8reader.readFile(envel,'envel')
    
    # load bdsim optics
    bdsinst = _pybdsim._General.CheckItsBDSAsciiData(bdsim)
    bdsopt  = _GetBDSIMOptics(bdsinst)

    # parameters required for calculating beam sizes, not written in mad8 output so have to supply manually.
    beamParams = {'esprd': energySpread, 'particle': particle, 'ex': ex, 'ey':ey}

    # make plots 
    mad8opt = {'comm':com, 'twiss':twissL, 'envel':envelL, 'beam': beamParams}

    # energy and npart plotted with individual methods
    figures = [PlotBeta(mad8opt,bdsopt,functions=functions,
                         postfunctions=postfunctions,
                         figsize=figsize, xlim=xlim, survey=survey),
               PlotAlpha(mad8opt,bdsopt,functions=functions,
                          postfunctions=postfunctions,
                          figsize=figsize, xlim=xlim, survey=survey),
               PlotDisp(mad8opt,bdsopt,functions=functions,
                      postfunctions=postfunctions,
                      figsize=figsize, xlim=xlim, survey=survey),
               PlotDispP(mad8opt,bdsopt,functions=functions,
                       postfunctions=postfunctions,
                       figsize=figsize, xlim=xlim, survey=survey),
               PlotSigma(mad8opt,bdsopt,functions=functions,
                          postfunctions=postfunctions,
                          figsize=figsize, xlim=xlim, survey=survey),
               PlotSigmaP(mad8opt,bdsopt,functions=functions,
                           postfunctions=postfunctions,
                           figsize=figsize, xlim=xlim, survey=survey),
               PlotEnergy(mad8opt,bdsopt,functions=functions,
                          postfunctions=postfunctions,
                          figsize=figsize, xlim=xlim, survey=survey),
               PlotMean(mad8opt,bdsopt,functions=functions,
                         postfunctions=postfunctions,
                         figsize=figsize, xlim=xlim, survey=survey),
               PlotEmitt(mad8opt,bdsopt,functions=functions,
                             postfunctions=postfunctions,
                             figsize=figsize, xlim=xlim, survey=survey),
               PlotNParticles(mad8opt,bdsopt,functions=functions,
                              postfunctions=postfunctions,
                              figsize=figsize, xlim=xlim, survey=survey)]
    if saveAll:
        tfsname = repr(twiss)
        bdsname = repr(bdsinst)
        output_filename = "optics-report.pdf"
        if outputFileName is not None:
            output_filename = outputFileName
            if not output_filename.endswith('.pdf'):
                output_filename += ".pdf"
        else:
            output_filename = fname.replace('.root','')
            output_filename += ".pdf"
        # Should have a more descriptive name really.
        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} (MAD8) VS {} (BDSIM) Optical Comparison".format(tfsname, bdsname)
            d['CreationDate'] = _datetime.datetime.today()

        print("Written ", output_filename)
    
    return mad8opt

def PlotEnergy(mad8opt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12,5), xlim=(0,0)) :
    N = str(int(bdsopt['Npart'][0]))  #number of primaries.
    energyPlot = _plt.figure('Energy',figsize)

    _plt.plot(mad8opt['twiss'].getColumn('suml'), # one missing energy due to initial 
              mad8opt['comm'].getColumn('E'),
              'b--', label=r'MAD8 $E$')

    _plt.errorbar(bdsopt['S'], bdsopt['Mean_E'],
                  yerr=bdsopt['Sigma_Mean_E'],
                  label=r'BDSIM $E$' + ' ; N = ' + N,
                  marker='x',
                  ls = '',
                  color='b')
    
    axes = _plt.gcf().gca()
    axes.set_ylabel('Energy / GeV')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    if survey is None:
        survey = mad8opt
    _CallUserFigureFunctions(functions)
    _AddSurvey(energyPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    if xlim != (0,0):
        _plt.xlim(xlim)

    energyPlot.sca(energyPlot.axes[0])
    
    _plt.show(block=False)
    return energyPlot

def PlotNParticles(mad8opt, bdsopt, survey=None, functions=None, postfunctions=None, figsize=(12, 5), xlim=(0,0)):
    npartPlot = _plt.figure('NParticles', figsize)

    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k-', label='BDSIM N Particles')
    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k.')
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'N Particles')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    if survey is None:
        survey = mad8opt
    _CallUserFigureFunctions(functions)
    _AddSurvey(npartPlot, survey)
    _CallUserFigureFunctions(postfunctions)
    if xlim != (0,0):
        _plt.xlim(xlim)

    npartPlot.sca(npartPlot.axes[0])

    _plt.show(block=False)
    return npartPlot
    
def _AddSurvey(figure, survey):
    if survey is None:
        return 
    elif survey[-5:] == ".root":
        _pybdsim.Plot.AddMachineLatticeFromSurveyToFigure(figure,survey)
    else:
        _pymad8.Plot.AddMachineLatticeToFigure(figure,survey)

def _CallUserFigureFunctions(functions):
    if isinstance(functions, list):
        for function in functions:
            if callable(function):
                function()
    elif callable(functions):
        functions()

def ChangeYlim(plotname,limits):
    """
    Change ylimit on data subfigure after survey has been added to the plot.
    Limits must be supplied as a tuple (ymin,ymax).
    """
    f = _plt.figure(plotname)
    ax = f.axes[0]
    ax.set_ylim(limits)

# ============================================================================
# Below is old
# ============================================================================
class Mad8Bdsim :
    def __init__(self, 
                 bdsimFileName = "output.pickle",
                 mad8TwissFileName   = "ebds1.twiss",
                 mad8EnvelopeFileName = "ebds1.envelope") : 
        # load bdsim data
        if bdsimFileName.find("pickle") != -1 :
            f = open(bdsimFileName) 
            self.bdsimData = _pkl.load(f)
            self.bdsimOptics = self.bdsimData['optics']
            f.close()
        elif bdsimFileName.find(".root") != -1 :
            import ROOT as _ROOT 
            import root_numpy as _root_numpy
            f = _ROOT.TFile(bdsimFileName)
            t = f.Get("optics")
            self.bdsimOptics = _root_numpy.tree2rec(t)
            


        # load mad8 data
        r = _pymad8.Mad8.OutputReader()    
        [self.mad8Comm,self.mad8Envel] = r.readFile(mad8EnvelopeFileName,"envel")
        [self.mad8Comm,self.mad8Twiss] = r.readFile(mad8TwissFileName,"twiss")

    def plotSigma(self) : 
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _plt.subplot(gs[1])
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Envel.getColumn('s11'))*1e6,"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Sigma_x']/1e-6,"x--",label="BDSIM")
        #_pl.xlim(0,2275)
        #_pl.ylim(0,3000)
        _pl.legend(loc=0)
        _pl.ylabel("$\\sigma_x$ [$\mu$m]")

        ax2 = _plt.subplot(gs[2])
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Envel.getColumn('s33'))*1e6,"+-")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Sigma_y']/1e-6,"x--")
        #_pl.xlim(0,2275)
        #_pl.ylim(0,100)
        _pl.ylabel("$\\sigma_y$ [$\mu$m]")
        _pl.xlabel("$S$ [m]")

        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)

        _pl.savefig("mad8bdsim_sigma.pdf")

        
    def plotSigmaPrim(self) : 
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _plt.subplot(gs[1])
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Envel.getColumn('s22'))*1e6,"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Sigma_xp']/1e-6,"x--",label="BDSIM")
        #_pl.xlim(0,2275)
        #_pl.ylim(0,3000)
        _pl.legend(loc=0)
        _pl.ylabel("$\\sigma^{'}_{x}$ [$\mu$m]")

        ax2 = _plt.subplot(gs[2])
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Envel.getColumn('s44'))*1e6,"+-")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Sigma_yp']/1e-6,"x--")
        #_pl.xlim(0,2275)
        #_pl.ylim(0,100)
        _pl.ylabel("$\\sigma^{'}_{y}$ [$\mu$m]")
        _pl.xlabel("$S$ [m]")

        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)

        _pl.savefig("mad8bdsim_sigma_prim.pdf")

    def plotOrbit(self) :
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _plt.subplot(gs[1])
        _pl.plot(self.mad8Envel.getColumn('suml'), _np.zeros(len(self.mad8Envel.getColumn('suml'))),"+-",label="MAD8") #mad8 orbit perfectly on reference
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Mean_x']/1e-6,"x--",label="BDSIM")
        #_pl.xlim(0,2275)
        #_pl.ylim(0,3000)
        _pl.legend(loc=0)
        _pl.ylabel("$\\overline{x}$ [$\mu$m]")

        ax2 = _plt.subplot(gs[2])
        _pl.plot(self.mad8Envel.getColumn('suml'), _np.zeros(len(self.mad8Envel.getColumn('suml'))) ,"+-")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Mean_y']/1e-6,"x--")
        _pl.xlim(0,2275)
        #_pl.ylim(0,100)
        _pl.ylabel("$\\overline{y}$ [$\mu$m]")
        _pl.xlabel("$S$ [m]")

        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)
        
        _pl.savefig("mad8bdsim_mean.pdf")

    
    def plotBeta(self) : 
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)
        
        ax1 = _pl.subplot(gs[1])
        ax1.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Twiss.getColumn('betx')),"+-")
        _pl.plot(self.bdsimOptics['S'],_pl.sqrt(self.bdsimOptics['Beta_x']),"+--")
        _pl.ylabel("$\sqrt\\beta_x$ [m]")

        ax2 = _pl.subplot(gs[2])
        ax2.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),_pl.sqrt(self.mad8Twiss.getColumn('bety')),"+-")
        _pl.plot(self.bdsimOptics['S'],_pl.sqrt(self.bdsimOptics['Beta_y']),"+--")
        _pl.ylabel("$\sqrt\\beta_y$ [m]")
        _pl.xlabel("$S$ [m]")

        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)
        
        _pl.savefig("mad8bdsim_beta.pdf")
    

    def plotSurvey(self, mad8SurveyFileName, bdsimSurveyFileName) :
        # load bdsim survey
        fs = _pybdsim.Data.Load(bdsimSurveyFileName)

        # load mad8 survey
        rs = _pymad8.Mad8.OutputReader()    
        [common, mad8Survey] = rs.readFile(mad8SurveyFileName,"survey")
        
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _pl.subplot(gs[1])
        _pl.plot(mad8Survey.getColumn('suml'), mad8Survey.getColumn('x'),"+-", label = "MAD8")
        _pl.plot(fs.SStart(),fs.X(),"+--", label = "BDSIM")
        #_pl.xlim(0,max(mad8Survey.getColumn('suml')))
        _pl.ylabel("$X$ [m]")

        ax2 = _pl.subplot(gs[2])
        _pl.plot(mad8Survey.getColumn('suml'),mad8Survey.getColumn('y'),"+-", label = "MAD8")
        _pl.plot(fs.SStart(),fs.Y(),"+--", label = "BDSIM")
        #_pl.xlim(0,max(mad8Survey.getColumn('suml')))
        _pl.ylabel("$Y$ [m]")
        _pl.xlabel("$S$ [m]")

        _pl.legend(loc=0)
        _pl.subplots_adjust(hspace=0.25,top=0.94,left=0.1,right=0.92)

        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)
        
        _pl.savefig("mad8bdsim_survey.pdf")


    def plotDispersion(self) :
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _pl.subplot(gs[1])
        ax1.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),self.mad8Twiss.getColumn('dx'),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Disp_x'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\eta_x$ [m]")

        ax2 = _pl.subplot(gs[2])
        ax2.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),self.mad8Twiss.getColumn('dy'),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Disp_y'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\eta_y$ [m]")
        _pl.xlabel("$S$ [m]")
        
        _pl.legend(loc=0)
        
        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)

        _pl.savefig("mad8bdsim_eta.pdf")
        

    def plotDispersionPrim(self) :
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _pl.subplot(gs[1])
        ax1.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),self.mad8Twiss.getColumn('dpx'),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Disp_xp'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\eta^{'}_{x}$ [rad]")

        ax2 = _pl.subplot(gs[2])
        ax2.set_autoscale_on(True)
        _pl.plot(self.mad8Envel.getColumn('suml'),self.mad8Twiss.getColumn('dpy'),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Disp_yp'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\eta^{'}_{y}$ [rad]")
        _pl.xlabel("$S$ [m]")
        
        _pl.legend(loc=0)
        
        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)

        _pl.savefig("mad8bdsim_etaprim.pdf")

        
    def plotEmittance(self) :
        figure = _plt.figure(figsize=(11.6, 7.2))
        gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
        ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
        _pymad8.Plot.drawMachineLattice(self.mad8Comm,self.mad8Twiss)

        ax1 = _pl.subplot(gs[1])
        ax1.set_autoscale_on(True)
        #_pl.plot(self.mad8Envel.getColumn('suml'), _np.zeros(len(self.mad8Envel.getColumn('suml'))),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Emitt_x'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\epsilon_{x}$ [m]")

        ax2 = _pl.subplot(gs[2])
        ax2.set_autoscale_on(True)
        #_pl.plot(self.mad8Envel.getColumn('suml'), _np.zeros(len(self.mad8Envel.getColumn('suml'))),"+-",label="MAD8")
        _pl.plot(self.bdsimOptics['S'],self.bdsimOptics['Emitt_y'],"+--",label="BDSIM") # 4000 1/250*1e6
        _pl.ylabel("$\\epsilon_{y}$ [m]")
        _pl.xlabel("$S$ [m]")
        
        _pl.legend(loc=0)
        
        _pymad8.Plot.setCallbacks(figure,ax0,[ax1,ax2],self.mad8Twiss)

        _pl.savefig("mad8bdsim_emitt.pdf")



"""Functions for generating the output files for regression testing as
well as some utility functions which are used both here and elsewhere
for easily accessing the outputs.  The starting point here is the
input file name, e.g. "atf2-nominal-twiss-v5.2.tfs.tar.gz", which is
the name of the input found in pybdsim/tests/test_input/.   """

import pybdsim
import pymadx

from . import pybdsim_test_utils as utils

def generate_atf2_gmad():
    input_file_name = "atf2-nominal-twiss-v5.2.tfs.tar.gz"
    input_path, output_path = utils.get_input_and_output_paths(
        input_file_name)
    pybdsim.Convert.MadxTfs2Gmad(input_path, output_path)

def generate_model_model_gmad():
    input_file_name = "model-model.tfs.gz"
    input_path, output_path = utils.get_input_and_output_paths(
        input_file_name)
    # Load the collimator settings
    collsettings_path = utils.get_input_path("model-model-collsettings.dat")
    coll_settings =  pybdsim.Data.Load(collsettings_path)
    # Load the aperture information
    aper_path = utils.get_input_path("model-model-aper.tfs.gz")
    aper = pymadx.Data.Aperture(aper_path)
    aper = aper.RemoveBelowValue(0.005)

    pybdsim.Convert.MadxTfs2Gmad(input_path, output_path,
                                 aperturedict=aper,
                                 collimatordict=coll_settings)

if __name__ == "__main__":
    generate_atf2_gmad()
    generate_model_model_gmad()



"""
Module containing a similarly named Beam class for creating
a BDSIM beam distribution programmatically.
"""

BDSIMDistributionTypes = [
    'reference',
    'gaussmatrix',
    'gauss',
    'gausstwiss',
    'circle',
    'square',
    'ring',
    'eshell',
    'halo',
    'composite',
    'userfile',
    'ptc',
    'sixtrack',
    'eventgeneratorfile',
    'sphere',
    'compositesde',
    'box',
    'bdsimsampler',
    'halosigma'
]

BDSIMParticleTypes = [
    'e-',
    'e+',
    'proton',
    'gamma',
    'antiproton',
    'pi+',
    'pi-',
    'neutron',
    'photon',
    'mu+',
    'mu-',
    'kaon-',
    'kaon+',
    'kaon0L'
]

class Beam(dict) :
    def __init__(self, particleType='e-', energy=1.0, distrType='reference', *args, **kwargs):
        dict.__init__(self,*args,**kwargs)
        self.SetParticleType(particleType)
        self.SetEnergy(energy)
        self.SetDistributionType(distrType)
        
    def SetParticleType(self,particleType='e-'):
        if particleType not in BDSIMParticleTypes:
            raise ValueError("Unknown particle type: '"+str(particleType)+"'")
        self['particle'] = '"' + str(particleType) + '"'

    def SetEnergy(self,energy=1.0,unitsstring='GeV'):
        self['energy'] = str(energy) + '*' + unitsstring

    def _MakeGaussTwiss(self):
        setattr(self, 'SetBetaX',      self._SetBetaX)
        setattr(self, 'SetBetaY',      self._SetBetaY)
        setattr(self, 'SetAlphaX',     self._SetAlphaX)
        setattr(self, 'SetAlphaY',     self._SetAlphaY)
        setattr(self, 'SetEmittanceX', self._SetEmittanceX)
        setattr(self, 'SetEmittanceY', self._SetEmittanceY)
        setattr(self, 'SetEmittanceNX', self._SetEmittanceNX)
        setattr(self, 'SetEmittanceNY', self._SetEmittanceNY)
        setattr(self, 'SetSigmaE',     self._SetSigmaE)
        setattr(self, 'SetSigmaT',     self._SetSigmaT)
        setattr(self, 'SetDispX',      self._SetDispX)
        setattr(self, 'SetDispY',      self._SetDispY)
        setattr(self, 'SetDispXP',     self._SetDispXP)
        setattr(self, 'SetDispYP',     self._SetDispYP)

    def SetDistributionType(self,distrType='reference'):
        if distrType not in BDSIMDistributionTypes:
            raise ValueError("Unknown distribution type: '"+str(distrType)+"'")
        self['distrType'] = '"' + distrType + '"'
        self._UpdateMemberFunctions(distrType)

    def __repr__(self):
        return self.ReturnBeamString()

    def _UpdateMemberFunctions(self, distrType):
        if distrType == 'reference':
            pass
        elif distrType == 'gaussmatrix':
            setattr(self, 'SegSigmaNM',    self._SetSigmaNM)
        elif distrType == 'gauss':
            setattr(self, 'SetSigmaX',     self._SetSigmaX)
            setattr(self, 'SetSigmaY',     self._SetSigmaY)
            setattr(self, 'SetSigmaE',     self._SetSigmaE)
            setattr(self, 'SetSigmaXP',    self._SetSigmaXP)
            setattr(self, 'SetSigmaYP',    self._SetSigmaYP)
            setattr(self, 'SetSigmaT',     self._SetSigmaT)   
        elif distrType == 'gausstwiss':
            self._MakeGaussTwiss()
        elif distrType == 'circle':
            setattr(self, 'SetEnvelopeR',  self._SetEnvelopeR)
            setattr(self, 'SetEnvelopeRp', self._SetEnvelopeRp)
            setattr(self, 'SetEnvelopeT',  self._SetEnvelopeT)
            setattr(self, 'SetEnvelopeE',  self._SetEnvelopeE)
        elif distrType == 'square':
            setattr(self, 'SetEnvelopeX',  self._SetEnvelopeX)
            setattr(self, 'SetEnvelopeXp', self._SetEnvelopeXp)
            setattr(self, 'SetEnvelopeY',  self._SetEnvelopeY)
            setattr(self, 'SetEnvelopeYp', self._SetEnvelopeYp)
            setattr(self, 'SetEnvelopeT',  self._SetEnvelopeT)
            setattr(self, 'SetEnvelopeE',  self._SetEnvelopeE)
        elif distrType == 'ring':
            setattr(self, 'SetRMin',       self._SetRMin)
            setattr(self, 'SetRMax',       self._SetRMax)
        elif distrType == 'eshell':
            setattr(self, 'SetShellX',     self._SetShellX)
            setattr(self, 'SetShellY',     self._SetShellY)
            setattr(self, 'SetShellXP',    self._SetShellXP)
            setattr(self, 'SetShellYP',    self._SetShellYP)
        elif distrType == 'halo':
            self._MakeGaussTwiss()
            setattr(self, 'SetHaloNSigmaXInner',      self._SetHaloNSigmaXInner)
            setattr(self, 'SetHaloNSigmaXOuter',      self._SetHaloNSigmaXOuter)
            setattr(self, 'SetHaloNSigmaYInner',      self._SetHaloNSigmaYInner)
            setattr(self, 'SetHaloNSigmaYOuter',      self._SetHaloNSigmaYOuter)
            setattr(self, 'SetHaloPSWeightParameter', self._SetHaloPSWeightParameter)
            setattr(self, 'SetHaloPSWeightFunction',  self._SetHaloPSWeightFunction)
            setattr(self, 'SetHaloXCutInner',         self._SetHaloXCutInner)
            setattr(self, 'SetHaloYCutInner',         self._SetHaloYCutInner)
        elif distrType == 'composite':
            setattr(self, 'SetXDistrType', self._SetXDistrType)
            setattr(self, 'SetYDistrType', self._SetYDistrType)
            setattr(self, 'SetZDistrType', self._SetZDistrType)
        elif distrType == 'ptc' : 
            setattr(self, 'SetSigmaE',     self._SetSigmaE)            
            setattr(self, 'SetDistribFileName',self._SetDistribFileName)
        elif distrType == "userfile":
            setattr(self, 'SetDistrFile',     self._SetDistrFile)
            setattr(self, 'SetDistrFileFormat',self._SetDistrFileFormat)

    def WriteToFile(self, filename):
        f = open(filename, 'w')
        f.write(self.ReturnBeamString())
        f.close()
            
    def ReturnBeamString(self):
        s = ''
        for k,v in sorted(self.items()):
            s += ', \n\t'+str(k)+'='+str(v)
        s += ';'
        s2 = s.split('\n')
        s3 = 'beam,\t'+s2[1].replace('\t','').replace('\n','').replace(',','').strip()+',\n'
        s4 = '\n'.join(s2[2:])
        st = s3+s4
        return st

    def SetX0(self,x0=0.0,unitsstring='m'):
        self['X0'] = str(x0) + '*' + unitsstring

    def SetY0(self,y0=0.0,unitsstring='m'):
        self['Y0'] = str(y0) + '*' + unitsstring

    def SetZ0(self,z0=0.0,unitsstring='m'):
        self['Z0'] = str(z0) + '*' + unitsstring

    def SetXP0(self,xp0=0.0):
        self['Xp0'] = str(xp0)

    def SetYP0(self,yp0=0.0):
        self['Yp0'] = str(yp0)

    def SetZP0(self,zp0=0.0):
        self['Zp0'] = str(zp0)

    def SetS0(self, s0=0,unitsstring='m'):
        self['S0'] = str(s0) + '*' + unitsstring

    def SetE0(self, e0=1, unitsstring='GeV'):
        self['E0'] = str(e0) + '*' + unitsstring

    def SetT0(self,t0=0.0,unitsstring='s'):
        self['T0'] = str(t0) + '*' + unitsstring

    def _SetSigmaNM(self, n, m, value):
        self['sigma'+str(n)+str(m)] = value

    def _SetSigmaX(self,sigmax=1.0,unitsstring='um'):
        self['sigmaX'] = str(sigmax) + '*' + unitsstring

    def _SetSigmaY(self,sigmay=1.0,unitsstring='um'):
        self['sigmaY'] = str(sigmay) + '*' + unitsstring

    def _SetSigmaE(self,sigmae=0.001):
        """
        fractional energy spread
        """
        self['sigmaE'] = str(sigmae)

    def _SetSigmaXP(self,sigmaxp=1.0,unitsstring='mrad'):
        self['sigmaXp'] = str(sigmaxp) + '*' + unitsstring

    def _SetSigmaYP(self,sigmayp=1.0,unitsstring='mrad'):
        self['sigmaYp'] = str(sigmayp) + '*' + unitsstring

    def _SetSigmaT(self,sigmat=1.0,unitsstring='s'):
        self['sigmaT'] = str(sigmat)

    def _SetBetaX(self, betx=1.0, unitsstring='m'):
        self['betx'] = str(betx) + '*' + unitsstring

    def _SetBetaY(self,bety=1.0,unitsstring='m'):
        self['bety'] = str(bety) + '*' + unitsstring

    def _SetAlphaX(self,alphax=1.0,unitsstring='m'):
        self['alfx'] = str(alphax)

    def _SetAlphaY(self,alphay=1.0,unitsstring='m'):
        self['alfy'] = str(alphay)

    def _SetDispX(self,dispx=1.0,unitsstring='m'):
        self['dispx'] = str(dispx) + '*' + unitsstring

    def _SetDispY(self,dispy=1.0,unitsstring='m'):
        self['dispy'] = str(dispy) + '*' + unitsstring

    def _SetDispXP(self,dispxp=1.0):
        self['dispxp'] = str(dispxp)

    def _SetDispYP(self,dispyp=1.0):
        self['dispyp'] = str(dispyp)

    def _SetEmittanceX(self,emitx=1.0e-9,unitsstring='um'):
        self['emitx'] = str(emitx) + '*' + unitsstring
   
    def _SetEmittanceY(self,emity=1.0e-9,unitsstring='um'):
        self['emity'] = str(emity) + '*' + unitsstring

    def _SetEmittanceNX(self,emitnx=1.0e-9,unitsstring='mm*mrad'):
        self['emitnx'] = str(emitnx) + '*' + unitsstring

    def _SetEmittanceNY(self,emitny=1.0e-9,unitsstring='mm*mrad'):
        self['emitny'] = str(emitny) + '*' + unitsstring

    def _SetShellX(self,shellx=1.0,unitsstring='m'):
        self['shellX'] = str(shellx) + '*' + unitsstring

    def _SetShellY(self,shelly=1.0,unitsstring='m'):
        self['shellY'] = str(shelly) + '*' + unitsstring

    def _SetShellXP(self,shellxp=1.0):
        self['shellXp'] = str(shellxp)

    def _SetShellYP(self,shellyp=1.0):
        self['shellYp'] = str(shellyp)

    def _SetEnvelopeR(self, enveloper=1.0, unitsstring='um'):
        self['envelopeX'] = str(enveloper) + '*' + unitstring

    def _SetEnvelopeRp(self, enveloperp=1.0):
        self['envelopeY'] = str(enveloperp) + '*' + unitstring

    def _SetEnvelopeT(self, envelopet=1.0,unitstring='s'):
        self['envelopeX'] = str(envelopet) + '*' + unitstring

    def _SetEnvelopeE(self, envelopee=1.0,unitstring='GeV'):
        self['envelopeY'] = str(envelopee) + '*' + unitstring

    def _SetEnvelopeX(self, envelopex=1.0,unitstring='m'):
        self['envelopeX'] = str(envelopex) + '*' + unitstring

    def _SetEnvelopeY(self, envelopey=1.0,unitstring='m'):
        self['envelopeY'] = str(envelopey) + '*' + unitstring

    def _SetEnvelopeXp(self, envelopexp=1.0):
        self['envelopeXp'] = str(envelopexp) + '*' + unitstring

    def _SetEnvelopeYp(self, envelopeyp=1.0):
        self['envelopeYp'] = str(envelopeyp)

    def _SetHaloNSigmaXInner(self, halonsigmaxinner=1):
        self['haloNSigmaXInner'] = str(halonsigmaxinner)

    def _SetHaloNSigmaXOuter(self, halonsigmaxouter=2):
        self['haloNSigmaXOuter'] = str(halonsigmaxouter)

    def _SetHaloNSigmaYInner(self, halonsigmayinner=1):
        self['haloNSigmaYInner'] = str(halonsigmayinner)

    def _SetHaloNSigmaYOuter(self, halonsigmayouter=2):
        self['haloNSigmaYOuter'] = str(halonsigmayouter)

    def _SetHaloPSWeightParameter(self, param):
        self['haloPSWeightParameter'] = param

    def _SetHaloPSWeightFunction(self, func):
        self['haloPSWeightFunction'] = '"'+func+'"'

    def _SetHaloXCutInner(self, haloxcutinner=0):
        self['haloXCutInner'] = str(haloxcutinner)

    def _SetHaloYCutInner(self, haloycutinner=0):
        self['haloYCutInner'] = str(haloycutinner)

    def _SetRMin(self,rmin=0.9,unitsstring='mm'):
        self['Rmin'] = str(rmin) + '*' + unitsstring
    
    def _SetRMax(self,rmax=1.0,unitsstring='mm'):
        self['Rmax'] = str(rmax) + '*' + unitsstring

    def _SetDistribFileName(self, fileName) :
        self['distrFile'] = '"'+fileName+'"'

    def _SetXDistrType(self, name):
        self._UpdateMemberFunctions(name)
        self['xDistrType'] = '"' + name + '"'

    def _SetYDistrType(self, name):
        self._UpdateMemberFunctions(name)
        self['yDistrType'] = '"' + name + '"'

    def _SetZDistrType(self, name):
        self._UpdateMemberFunctions(name)
        self['zDistrType'] = '"' + name + '"'

    def _SetOffsetSampleMean(self,on=False):
        if on == True:
            self["offsetSampleMean"] = 1
        else:
            self["OffsetSampleMean"] = 0

    def _SetDistrFile(self, filename):
        self["distrFile"] = '"{}"'.format(filename)

    def _SetDistrFileFormat(self, format_string):
        self["distrFileFormat"] = '"{}"'.format(format_string)

    def _MakeGaussTwiss(self):
        setattr(self, 'SetBetaX',      self._SetBetaX)
        setattr(self, 'SetBetaY',      self._SetBetaY)
        setattr(self, 'SetAlphaX',     self._SetAlphaX)
        setattr(self, 'SetAlphaY',     self._SetAlphaY)
        setattr(self, 'SetEmittanceX', self._SetEmittanceX)
        setattr(self, 'SetEmittanceY', self._SetEmittanceY)
        setattr(self, 'SetEmittanceNX', self._SetEmittanceNX)
        setattr(self, 'SetEmittanceNY', self._SetEmittanceNY)
        setattr(self, 'SetSigmaE',     self._SetSigmaE)
        setattr(self, 'SetSigmaT',     self._SetSigmaT)
        setattr(self, 'SetDispX',      self._SetDispX)
        setattr(self, 'SetDispY',      self._SetDispY)
        setattr(self, 'SetDispXP',     self._SetDispXP)
        setattr(self, 'SetDispYP',     self._SetDispYP)



#explicit imports to keep namespace clean

from ._MadxBdsimComparison import MadxVsBDSIM
from ._MadxBdsimComparison import MadxVsBDSIMOrbit
from ._MadxBdsimComparison import MadxVsBDSIMOrbitResiduals
from ._MadxBdsimComparison import MadxVsBDSIMFromGMAD
from ._BdsimBdsimComparison import BDSIMVsBDSIM
from ._BdsimBdsimComparison import PTCVsBDSIM
from ._TransportBdsimComparison import TransportVsBDSIM
from ._MadxMadxComparison import MadxVsMadx
from ._MultipleCodeComparison import Optics
from ._MultipleCodeComparison import OpticsResiduals
from ._ElegantBdsimComparison import ElegantVsBDSIM

# optional pymad8
try:
    from ._Mad8BdsimComparison import Mad8VsBDSIM
except ImportError:
    pass

# optional pysad
try:
    import pysad as _pysad
    from ._SadComparison import SadComparison
except ImportError:
    pass


